/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Parser trs simplifi pour lire les fichiers INI
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "defs.h"

#ifdef _GNU_SOURCE
#   define PORTFILE "/etc/cc90/cc90.ini"
#else
#   define PORTFILE "c:\\windows\\cc90\\cc90.ini"
#endif

extern struct CC90_T cc90 ;
extern struct OPT_T opt ;
extern struct CONV_T conv ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
void FreeFile(void) ;

static char linebuffer[STRING_SIZE_MAX] ; /* Buffer de ligne */
static char *line ;           /* Pointeur de travail sur linebuffer */
static UBYTE /*@null@*//*@out@*/ *inibuf = NULL ;  /* Pointeur de travail sur fbuffer */

extern int PError(char *, int) ;
extern int LoadFpFile(char *) ;
extern int InitConversionFormat(char *) ;

/*
 * Capture d'une ligne du fichier
 */
static int GetIniLine(void)
{
    if ((inibuf == NULL)
     || (inibuf >= (file_buffer + file_size)))
         return FALSE ;

    line = linebuffer ;

    while ((inibuf < (file_buffer + file_size))
        && (line < (linebuffer + STRING_SIZE_MAX - 1))
        && (*inibuf != 0x0A)
        && (*inibuf != 0x0D)
        && (*inibuf != 0x00))
        *(line++) = (char)*(inibuf++) ;

    while ((inibuf < (file_buffer + file_size))
        && (*inibuf != 0x0A)
        && (*inibuf != 0x0D)
        && (*inibuf != 0x00))
        inibuf++ ;

    /* Clture la ligne */
    *line = '\0' ;

    switch(*inibuf)
    {
        case 0x0A : if (*(inibuf+1) == 0x0D) inibuf++ ; break ;
        case 0x0D : if (*(inibuf+1) == 0x0A) inibuf++ ; break ;
    }
    inibuf++ ;
    line = linebuffer ;
    return TRUE ;
}


/*
 * Recherche la section
 */
static int FindIniSection(char *name)
{
    char section_name[strlen(name)+3] ;

    (void)snprintf(section_name,STRING_SIZE_MAX-1,"[%s]",name) ;

    /* Recherche le nom de la section */
    if (inibuf != NULL)
        while (GetIniLine() == TRUE)
            if (strncmp(section_name,line,strlen(section_name)) == 0)
                return TRUE ;
    return PError(name,ERR_SECTION_NOT_FOUND) ;
}


/*
 * Recherche la clef
 */
static int FindIniKey(char *name)
{
    /* Recherche le nom de la clef */
    while ((inibuf != NULL) && (GetIniLine() == TRUE))
    {
        if (*inibuf == (UBYTE)'[')
            return PError(name,ERR_KEY_NOT_FOUND) ;
        else
        if (strncmp(name,line,strlen(name)) == 0)
        {
            line += strlen(name) ;
            if (*(line++) != ' ')
                return PError(name,ERR_SPACE_AFTER_KEY) ;
            if (*(line++) != '=')
                return PError(name,ERR_EQUAL_AFTER_KEY) ;
            while((*line == ' ')
               || (*line == '\9'))
                line++ ;
            return TRUE ;
        }
    }
    return PError(name,ERR_KEY_NOT_FOUND) ;
}


/*
 * Recherche la chane de la valeur
 */
static int GetString(char *section_name, char *key_name, char *string, int maxsize)
{
    int i = 0 ;

    if (inibuf != NULL) { free(inibuf) ; inibuf = NULL ; }
    inibuf = file_buffer ;

    if (FindIniSection(section_name) == FALSE)
        return FALSE ;
    if (FindIniKey(key_name) == FALSE)
        return FALSE ;

    while((i < maxsize)
       && (*line != ';')
       && (*line > ' '))
        string[i++] = *(line++) ;
    string[i] = '\0' ;

    return TRUE ;
}


/*
 * Rcupre une valeur dcimale dans un INT
 */
static int GetInt(char *section_name, char *key_name, int *ptvalue)
{
    size_t index ;
    static char string[5] ;

    if (GetString(section_name,key_name,string,4) == FALSE)
        return FALSE ;
    for(index=0;index<strlen(string);index++)
        if (isdigit((int)string[index]) == 0)
            return PError(string,ERR_VALUE_NOT_DECIMAL) ;
    if (string[0] != '\0')
        *ptvalue = atoi(string) ;
    return TRUE ;
}


/*
 * Traite le fichier INI
 */
static int GetCc90Config(void)
{
    static char string[STRING_SIZE_MAX + 2] ;

    if (GetString("Cc90","port.name",cc90.portname,39) == FALSE)    /* Nom du port srie */
        return FALSE ;
    if (GetInt("Disk","interlace",&opt.interlace) == FALSE)  /* Entrelacement par dfaut */
        return FALSE ;
    if (GetString("Files","BMP",string,STRING_SIZE_MAX) == FALSE)  /* Conversion pour les fichiers BMP */
        return FALSE ;
    if (InitConversionFormat(string) == FALSE)
        return PError(string,ERR_VALUE_NOT_DECIMAL) ;
    return FALSE ;
}


/*
 * Charge et traite le fichier INI
 */
int LoadIni(void)
{
    int success ;
    if (LoadFpFile(PORTFILE) == FALSE) return FALSE ;
    success = GetCc90Config() ;
    FreeFile() ;

    return success ;
}

