/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Conversion de fichiers SND/SON<->WAV
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h> /*****************************/
#include <stdlib.h>
#include <string.h>

#include "defs.h"

extern struct CC90_T cc90 ;
extern struct CONV_T conv ;
extern struct DISK_T dk[4] ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern UBYTE *temp_buffer ;
extern ULONG temp_size ;
extern void AllocTemp(size_t) ;
extern void FreeTemp(size_t) ;


extern int PError(char *, int) ;
extern void ExtractBin(void) ;
extern int RawToBin(void) ;

#define WAVE_HEADER_SIZE   44

/*
 * Lit un INT little endian en mmoire
 */
static ULONG StrToULong(char *buf)
{
    return ((ULONG)*(buf) << 24)
         | ((ULONG)*(buf+1) << 16)
         | ((ULONG)*(buf+2) << 8)
         | (ULONG)*(buf+3) ;
}


/*
 * Lit un INT little endian en mmoire
 */
ULONG /*@unused@*/ ReadLELong(UBYTE /*@special@*/*buf)
{
    return (ULONG)((*(buf+3) << 24) | (*(buf+2) << 16) | (*(buf+1) << 8) | *buf) ;
}


/*
 * Lit un SHORT little endian en mmoire
 */
USHORT /*@unused@*/ ReadLEWord(UBYTE /*@special@*/*buf)
{
    return (USHORT)((*(buf+1) << 8) | *buf) ;
}


/*
 * Lit un INT big endian en mmoire
 */
static ULONG ReadBELong(UBYTE /*@special@*/*buf)
{
    return (ULONG)((*buf << 24) | (*(buf+1) << 16) | (*(buf+2) << 8) | *(buf+3))  ;
}


/*
 * Lit un SHORT big endian en mmoire
 */
USHORT ReadBEWord(UBYTE /*@special@*/*buf)
{
    return (USHORT)((*buf << 8) | *(buf+1)) ;
}


/*
 * Ecrit un INT little endian en mmoire
 */
void /*@unused@*/ WriteLELong(UBYTE /*@special@*/*buf, ULONG data)
{
    *(buf) = (UBYTE)data ;
    *(buf+1) = (UBYTE)(data >> 8) ;
    *(buf+2) = (UBYTE)(data >> 16) ;
    *(buf+3) = (UBYTE)(data >> 24) ;
}


/*
 * Ecrit un SHORT little endian en mmoire
 */
void /*@unused@*/ WriteLEWord(UBYTE /*@special@*/*buf, USHORT data)
{
    *(buf) = (UBYTE)data ;
    *(buf+1) = (UBYTE)(data >> 8) ;
}


/*
 * Ecrit un INT big endian en mmoire
 */
static void WriteBELong(UBYTE /*@special@*/*buf, ULONG data)
{
    *(buf) = (UBYTE)(data >> 24) ;
    *(buf+1) = (UBYTE)(data >> 16) ;
    *(buf+2) = (UBYTE)(data >> 8) ;
    *(buf+3) = (UBYTE)data ;
}


/*
 * Ecrit un SHORT big endian en mmoire
 */
void WriteBEWord(UBYTE /*@special@*/*buf, ULONG data)
{
    *(buf) = (UBYTE)(data >> 8) ;
    *(buf+1) = (UBYTE)data ;
}


/*
 * Recherche d'un chunk WAV
 */
static UBYTE /*@null@*//*@dependent@*/
*SearchWaveChunk(ULONG chunk_code)
{
    UBYTE /*@dependent@*/ *buf = file_buffer + 12 ;

    while((buf < (file_buffer + file_size)) && (ReadBELong(buf) != chunk_code))
        buf += ((ReadLELong(buf+4) + 1) & 0xFFFFFFFE) + 8 ;
    if (buf >= (file_buffer + file_size)) buf = NULL ;
    return buf ;
}


/*
 * Lecture d'un fichier WAV
 */
static int WavToSound(int sndtype)
{
    size_t index ;
    size_t shift ;
    UBYTE *src ;
    UBYTE *dst ;
    size_t srcsiz ;

    if (file_buffer == NULL)
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

    src = file_buffer ;

    /* Vrification de la validit du fichier */
    if ((ReadBELong(src) != StrToULong("RIFF"))        /* Vrifie if "RIFF" */
     || ((ReadLELong(src+4) + 8) != (ULONG)file_size)  /* Vrifie taille    */
     || (ReadBELong(src+8) != StrToULong("WAVE"))      /* Vrifie id "WAVE" */ 
     || ((src = (UBYTE/*@special@*/*)SearchWaveChunk(StrToULong("fmt "))) == NULL) /* Recherche id "fmt " */
     || (ReadLEWord(src+8) != 1)                        /* Vrifie compression */
     || (ReadLEWord(src+10) != 1)                       /* Vrifie nombre de canaux */
     || ((src = SearchWaveChunk(StrToULong("data"))) == NULL))  /* Recherche id "data" */
        return PError("",ERR_BAD_FILE_FORMAT) ;
        
    srcsiz = (size_t)ReadLELong(src+4) ;
    AllocTemp(srcsiz + 4) ;
    dst = temp_buffer ;
    src += 8 ;

    /* Dcale ventuellement les donnes vers la droite */
    shift = 0 ;
    index = 0 ;
    while((index<srcsiz) && ((src[index] & 0x40) == 0)) index++ ;
    if (index<srcsiz) shift = 1 ;
    index = 0 ;
    while((index<srcsiz) && ((src[index] & 0x80) == 0)) index++ ;
    if (index<srcsiz) shift = 2 ;
    if (shift != 0)
        for(index=0;index<srcsiz;index++) src[index] >>= shift ;

    if (sndtype == CONV_SND_SON)
    {
        /* Conversion du fichier SON */
        while ((src - file_buffer) < (int)srcsiz)
        {
            /* 00000011|11112222|22333333 */
            *(dst++) = ((*src << 2) & 0xfc) | ((*(src+1) >> 4) & 0x03) ;
            *(dst++) = ((*(src+1) << 4) & 0xf0) | ((*(src+2) >> 2) & 0x0f) ;
            *(dst++) = ((*(src+2) << 6) & 0xc0) | (*(src+3) & 0x3f) ;
            src += 4 ;
        }
     }
     else
     {
         memcpy(dst,src,srcsiz) ;
         dst += srcsiz ;
     }

    /* Officialisation des nouvelles adresse et taille de fichier */
    FreeTemp((size_t)(dst-temp_buffer)) ;

    /* Cre fichier binaire si fichier SND */
    if ((sndtype == CONV_SND_SND)
     && (RawToBin() == FALSE))
        return FALSE ;

    return TRUE ;
}


/*
 * Conversion d'un WAV en SND
 */
int WavToSnd(void)
{
    return WavToSound(CONV_SND_SND) ;
}


/*
 * Conversion d'un WAV en SON
 */
int WavToSon(void)
{
    return WavToSound(CONV_SND_SON) ;
}


/*
 * Ecriture d'un fichier WAV
 */
static int SoundToWav(int sndtype)
{
    USHORT shift ;    
    size_t size = ((file_size + 1) & 0xFFFFFFFE) ;
    UBYTE *src ;
    UBYTE *dst ;

    AllocTemp(WAVE_HEADER_SIZE + file_size + 4) ;
    if ((file_buffer == NULL) || (temp_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;
    src = file_buffer ;
    dst = temp_buffer ;

    if (sndtype == CONV_SND_SON)
        size = (file_size * 4) / 3 ;
    else
        ExtractBin() ;

    WriteBELong(dst   ,StrToULong("RIFF")) ; /**** RIFF id */
    WriteLELong(dst+4 ,(ULONG)(WAVE_HEADER_SIZE + size-8)) ;  /* RIFF size */
    WriteBELong(dst+8 ,StrToULong("WAVE")) ; /**** WAVE id */
    WriteBELong(dst+12,StrToULong("fmt ")) ; /**** fmt id */
    WriteLELong(dst+16,16) ;                    /* fmt size */
    WriteLEWord(dst+20,1) ;                     /* format */
    WriteLEWord(dst+22,1) ;                     /* channels number */
    WriteLELong(dst+24,conv.sample_rate) ;      /* bytes per second */
    WriteLELong(dst+28,conv.sample_rate) ;      /* bytes per sample */
    WriteLEWord(dst+32,1) ;                     /* block align */
    WriteLEWord(dst+34,8) ;                     /* significant bits per sample */
    WriteBELong(dst+36,StrToULong("data")) ; /**** data id */
    WriteLELong(dst+40,(ULONG)size) ;           /* data size */
    dst += WAVE_HEADER_SIZE ;

    /* Dcompresse le son */
    while ((size_t)(src - file_buffer) < file_size)
    {
        if (sndtype == CONV_SND_SON)
        {
            /* 00000011|11112222|22333333 */
            *(dst++) = (*src >> 2) & 0x3F ;
            *(dst++) = ((*src << 4) & 0x30) | ((*(src+1) >> 4) & 0x0F) ;
            *(dst++) = ((*(src+1) << 2) & 0x3C)  | ((*(src+2) >> 6) & 0x03) ;
            src += 2 ;
        }
        *(dst++) = *(src++) & 0x3F ;
    }

    /* Dcale ventuellement les donnes vers la gauche */
    shift = 2 ;
    src = temp_buffer + WAVE_HEADER_SIZE ;
    while((src<dst) && ((*src & 0x40) == 0)) src++ ;
    if (src<dst) shift = 1 ;
    src = temp_buffer + WAVE_HEADER_SIZE ;
    while((src<dst) && ((*src & 0x80) == 0)) src++ ;
    if (src<dst) shift = 0 ;
    if (shift == 0)
        for(src=temp_buffer+WAVE_HEADER_SIZE;src<dst;src++) *src <<= shift ;

    /* Word Align */
    if (((dst - temp_buffer) & 1) != 0)
        *(dst++) = 0x00 ;

    /* Officialisation des nouvelles adresse et taille de fichier */
    FreeTemp((size_t)(dst - temp_buffer)) ;

    return TRUE ;
}

/*
 * Conversion d'un SND en WAV
 */
int SndToWav(void)
{
    return SoundToWav(CONV_SND_SND) ;
}


/*
 * Conversion d'un SON en WAV
 */
int SonToWav(void)
{
    return SoundToWav(CONV_SND_SON) ;
}

