/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Primitives disquettes
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#include "defs.h"

#ifdef DEBUG_MODE
    extern FILE *fp_log ;
#endif

#define SAP_HEADER_SIZE  66
#define SAP_SECT_SIZE   262
#define SAP_MAGIC_NUM  0xB3

#define DATASIZE     (uint16_t)1024

enum {
    OP_READ,
    OP_WRITE
} ;

extern UBYTE *dos_buffer ;
extern UBYTE *track_buffer ;

extern int InitTrack20(int, int) ;
extern int PError(char *, int) ;

extern int SerialOpen(char *, long) ;
extern int SerialWrite(UBYTE *, size_t) ;
extern int SerialRead(UBYTE *, size_t) ;

extern struct CC90_T cc90 ;
extern struct OPT_T opt ;
extern struct DISK_T dk[4] ;

extern FILE *fp_file ;
extern int OpenFpFile(int, char *, char *) ;
extern void CloseFpFile(void) ;

extern int AskRequest(int, char *) ;


const struct CONTROLLER cont[8] = {
    { 1, 0,   51200, "qd", 1, 25,  50, 128 },
    { 2, 0,   51200, "qd", 1, 25,  50, 128 },
    { 3, 2,   81920, "fd", 1, 40,  80, 128 },
    { 4, 0,  163840, "fd", 1, 40,  80, 256 },
    { 0, 1,  327680, "fd", 1, 80, 160, 256 },
    { 0, 0,  655360, "fd", 2, 80, 160, 256 },
    { 0, 0,  983040, "fd", 3, 80, 160, 256 },
    { 0, 0, 1310720, "fd", 4, 80, 160, 256 }
};
#define CONTSIZE  8


static const char SapString[] = "\1SYSTEME D'ARCHIVAGE PUKALL S.A.P. (c) Alexandre PUKALL Avril 1998\0" ;

/*
 * Calcul des CRC secteur pour les fichiers SAP
 * Routine emprunte  la librarie libsap.c de TEO
 */
static USHORT crcpuk_temp;
static USHORT puktable[] = {
     0x0000, 0x1081, 0x2102, 0x3183,
     0x4204, 0x5285, 0x6306, 0x7387,
     0x8408, 0x9489, 0xa50a, 0xb58b,
     0xc60c, 0xd68d, 0xe70e, 0xf78f
  };

static void crc_pukall(USHORT c)
{
    USHORT index;

    index = (crcpuk_temp ^ c) & 0xf;
    crcpuk_temp = ((crcpuk_temp>>4) & 0xfff) ^ puktable[index];

    c >>= 4;

    index = (crcpuk_temp ^ c) & 0xf;
    crcpuk_temp = ((crcpuk_temp>>4) & 0xfff) ^ puktable[index];
}


/* ========================================================================== *
 *                           CREATION DE DISQUE
 * ========================================================================== */


/*
 * Code le secteur SAP
 */
static void DoMagic(UBYTE *buf, size_t size)
{
    size_t index ;
    for(index=0;index<size;index++)
        buf[index] ^= SAP_MAGIC_NUM ;
}


/*
 * Calcule le CRC du secteur SAP
 */
static void GetCrcPuk(UBYTE *buf, size_t size)
{
    size_t index ;
    crcpuk_temp = 0xffff ;
    for(index=0;index<size;index++)
        crc_pukall(buf[index]) ;
}


/*
 * Cre un secteur SAP pour criture
 */
static void BuildSapSector(UBYTE sector, UBYTE *buf, size_t size)
{
    /* Copie le secteur de la piste */
    memcpy(&buf[4], &track_buffer[(sector - 1) * size], size) ;

    /* Met le CRC  jour */
    GetCrcPuk(buf, size+4) ;
    DoMagic(&buf[4], size) ;
    buf[size+4] = (UBYTE)(crcpuk_temp >> 8) ;
    buf[size+5] = (UBYTE)crcpuk_temp ;
}


/*
 * Cration du disque SAP
 */
static int
SapCreate(int drive, int controller) /*@globals errno@*/
{
    UBYTE track, sector ;
    UBYTE buf[MAXSECSIZE + 6] ;

    buf[0] = 0x00 ;
    buf[1] = 0x00 ;
    memset(track_buffer, 0xE5, TRACKSIZE) ;

    if (cont[controller].sap != 0)
    {
        if (OpenFpFile(FALSE, dk[drive].dkname,"wb") == TRUE)
        {
            (void)fwrite(&cont[controller].sap, 1, 1, fp_file) ;
            (void)fwrite(&SapString[1], 1, strlen(SapString)-1, fp_file) ;

            for(track=0;track<cont[controller].nbtracks;track++)
            {
                for(sector=FIRSTB1;sector<=LASTB2;sector++)
                {
                    buf[2] = track ;
                    buf[3] = sector ;
                    BuildSapSector(sector,buf,cont[controller].secsize) ;
                    if (fwrite(buf, 1, cont[controller].secsize+6, fp_file) != cont[controller].secsize+6)
                        return PError(dk[drive].arg,ERR_IO_ERROR) ;
                }
            }
            CloseFpFile() ;
            if (InitTrack20(drive, controller) == FALSE)
                return FALSE ;
        }
    }
    return TRUE ;
}


/*
 * Cration du disque DCMOTO
 */
static int
DcCreate(int drive, int controller) /*@globals errno@*/
{
    UBYTE maxdrive ;
    UBYTE track ;
    size_t tracksize = cont[controller].secsize*SECTORS_PER_TRACK ;
    memset(track_buffer, 0xE5, TRACKSIZE) ;

    if (OpenFpFile(FALSE, dk[drive].dkname,"wb") == TRUE)
    {
        maxdrive = cont[controller].maxdrive-1 ;
        for(dk[drive].drive=0;dk[drive].drive<=maxdrive;dk[drive].drive++)
        {
            for(track=0;track<cont[controller].nbtracks;track++)
                if (fwrite(track_buffer, 1, tracksize, fp_file) != tracksize)
                    return PError(dk[drive].arg,ERR_IO_ERROR) ;
        }
        CloseFpFile() ;
        for(dk[drive].drive=0;dk[drive].drive<=maxdrive;dk[drive].drive++)
            if (InitTrack20(drive, controller) == FALSE)
                return FALSE ;
        dk[drive].drive = maxdrive ;
    }
    return TRUE ;
}


/*
 * Cre un disque virtuel
 */
int /*@unused@*/
Create(int drive, int controller) /*@globals errno@*/
{
    switch(OpenFpFile(TRUE, dk[drive].dkname,"rb"))
    {
        case FALSE :
            return FALSE ;
        case TRUE :
            CloseFpFile() ;
            if (AskRequest(OVERWRITE_FLAG,dk[drive].dkname) == FALSE)
                return FALSE ;
            break ;
    }

    switch(dk[drive].type)
    {
        case SAP_TYPE :
            if (SapCreate(drive, controller) == FALSE)
                return FALSE ;
            break ;
        case DCMOTO_TYPE :
            if (DcCreate(drive, controller) == FALSE)
                return FALSE ;
            break ;
    }
    return TRUE ;
}


/*
 * Ouverture et cration du disque cible si inexistant
 */
static int
CreateIfNotExist(int drive) /*@globals errno@*/
{
    int success ;
    if ((success = OpenFpFile(TRUE, dk[drive].dkname,"rb")) != TRUE)
    {
        if (success != FALSE)
        {
            if((drive & 0x01) == 0)
                return PError(dk[drive].dkname,ERR_FILE_NOT_FOUND) ;
            else
                if (Create(drive,dk[drive-1].cont) == FALSE)
                    return FALSE ;
            return TRUE ;
        }
    }
    return success ;
}


/* ========================================================================== *
 *                            OPERATIONS THOMSON
 * ========================================================================== */

/*
 * Calcul du CRC srie
 */
static USHORT MakeCrc(UBYTE *buf, size_t size)
{
    size_t index ;
    UBYTE crc_high = 0xe2,
          crc_low = 0x95,
          c ;

    /* Calcule le CRC pour toute la zone */
    for(index=0;index<size;index++)
    {
        c = crc_high ^ buf[index] ;
        c ^= (c >> 4) ;
        crc_low ^= (c >> 3) ;
        crc_high = (c << 4) ^ crc_low ;
        crc_low = (c << 5) ^ c ;
    }
    return (crc_high << 8) | crc_low ;
}


/*
 * Emission d'un bloc d'info par le port srie
 */
int /*@unused@*/ SendInfo(UBYTE code, UBYTE info1, UBYTE info2, UBYTE info3, UBYTE info4, UBYTE info5)
{
    USHORT crc ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"SendInfo = ") ;
#endif

    /* Cre le bloc info */
    cc90.info.code    = code  ;
    cc90.info.data[0] = info1 ;
    cc90.info.data[1] = info2 ;
    cc90.info.data[2] = info3 ;
    cc90.info.data[3] = info4 ;
    cc90.info.data[4] = info5 ;

    /* Met  jour le CRC du header */
    crc = MakeCrc(&cc90.info.code,6) ;
    cc90.info.crchigh = crc >> 8  ;
    cc90.info.crclow = crc & 0xFF ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"%d [%02x,%02x,%02x,%02x,%02x] crc=0x%02x%02x\n",
                cc90.info.code,
                (unsigned char)cc90.info.data[0],
                (unsigned char)cc90.info.data[1],
                (unsigned char)cc90.info.data[2],
                (unsigned char)cc90.info.data[3],
                (unsigned char)cc90.info.data[4],
                (unsigned char)cc90.info.crchigh,
                (unsigned char)cc90.info.crclow
                ) ;
#endif

    /* Envoie le bloc d'info */
    if (SerialWrite(&cc90.info.code,8) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;

    /* Lit les 2 codes d'erreur
     * Cette lecture permet de reprer si une erreur
     * est survenue ct Thomson, le Thomson envoyant
     * un double code d'erreur au lieu d'un double 0
     * Qui plus est, cette lecture est indispensable
     * si le terminal tourne sous Windows
     * En effet, si une erreur survient ct Thomson, le
     * programme Thomson attend d'abord que le terminal
     * se dconnecte
     * Le buffer de Windows, du fait de la vitesse de
     * transmission, est constamment plein, donc le
     * Thomson risquerait, si cette lecture n'existait
     * pas, d'attendre longtemps avant que la ligne se
     * libre */
    if (SerialRead(&cc90.synchro[0],2) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;
    if ((cc90.synchro[0] | cc90.synchro[1]) != 0)
        return PError("THOMSON",ERR_BREAK) ;
    return TRUE ;
}


/*
 * Emission d'un bloc de donnes par le port srie
 */
static int SendDatas(UBYTE *buf, size_t size)
{
    USHORT crc ;
    UBYTE tempbuf[DATASIZE+2] ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"SendDatas = ") ;
#endif

    /* Cre la zone de donnes du bloc */
    memcpy(tempbuf, buf, size) ;
    crc = MakeCrc(tempbuf, size) ;    /* Calcule le CRC de donnes */
    *(tempbuf+size)   = (UBYTE)(crc >> 8) ;
    *(tempbuf+size+1) = (UBYTE)crc ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"%d crc=0x%02x%02x\n",
                   size,
                   tempbuf[size],
                   tempbuf[size+1]) ;
#endif

    /* Envoie le bloc de donnes */
    if (SerialWrite(tempbuf,size+2) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;

    /* Lit les 2 codes d'erreur
     * Cette lecture permet de reprer si une erreur
     * est survenue ct Thomson, le Thomson envoyant
     * un double code d'erreur au lieu d'un double 0
     * Qui plus est, cette lecture est indispensable
     * si le terminal tourne sous Windows
     * En effet, si une erreur survient ct Thomson, le
     * programme Thomson attend d'abord que le terminal
     * se dconnecte
     * Le buffer de Windows, du fait de la vitesse de
     * transmission, est constamment plein, donc le
     * Thomson risquerait, si cette lecture n'existait
     * pas, d'attendre longtemps avant que la ligne se
     * libre */
    if (SerialRead(&cc90.synchro[0],2) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;
    if ((cc90.synchro[0] | cc90.synchro[1]) != 0)
        return PError("THOMSON",ERR_BREAK) ;

    return TRUE ;
}


/*
 * Rception d'un bloc d'info par le port srie
 */
static int GetInfo(void)
{
#ifdef DEBUG_MODE
    fprintf(fp_log,"GetInfo = ") ;
#endif

    /* Lit le bloc d'info */
    if (SerialRead(&cc90.info.code,8) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"%d [%02x,%02x,%02x,%02x,%02x] crc=0x%02x%02x\n",
                cc90.info.code,
                (unsigned char)cc90.info.data[0],
                (unsigned char)cc90.info.data[1],
                (unsigned char)cc90.info.data[2],
                (unsigned char)cc90.info.data[3],
                (unsigned char)cc90.info.data[4],
                (unsigned char)cc90.info.crchigh,
                (unsigned char)cc90.info.crclow) ;
#endif

    /* Vrifie le CRC d'info */
    if ((USHORT)((cc90.info.crchigh << 8) | cc90.info.crclow) != MakeCrc(&cc90.info.code,6))
        return PError(cc90.portname,ERR_CRC_ERROR) ;

    return TRUE ;
}


/*
 * Rception d'un bloc de donnes par le port srie
 */
static int GetDatas(UBYTE *buf, size_t size)
{
    UBYTE /*@out@*/ tempbuf[DATASIZE+2] ;

    memset(tempbuf,0x00,DATASIZE+2) ;
#ifdef DEBUG_MODE
    fprintf(fp_log,"GetDatas = ") ;
#endif

    /* Lit le bloc de donnes */
    if (SerialRead((UBYTE /*@out@*/*)tempbuf,size+2) == FALSE)
        return PError(cc90.portname,ERR_SERIAL_ERROR) ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"%d\n",size) ;
#endif

    /* Vrifie le CRC de donnes */
    if ((USHORT)((tempbuf[size] << 8) | (tempbuf[size+1] & 0xFF)) != MakeCrc(tempbuf,size))
        return PError(cc90.portname,ERR_CRC_ERROR) ;

    /* Copie le bloc de donnes */
    memcpy(buf, tempbuf, size) ;

    return TRUE ;
}


/*
 * Initialisation du drive Tomson
 */
static int ToInit(int drive)
{
    time_t x;
    struct tm *t;

    int i = 0 ;

    if (SerialOpen(cc90.portname,cc90.timeout) == FALSE)
        return PError(cc90.portname,ERR_PORT_NOT_READY) ;

    /* Envoie un bloc ASK au Thomson */
    (void)time(&x);
    t = gmtime(&x);
    if (t == NULL) return FALSE ;
    
    if (SendInfo(CODE_ASK,
                    (UBYTE)(((dk[drive].external == TRUE) ? 0x80 : 0x00)
                      | (((cc90.diskop == TRUE) && (dk[drive].drive <= 3)) ? 0x40 : 0x00)
                      | dk[drive].drive),
                    ((UBYTE)t->tm_mday << 3) | (((UBYTE)t->tm_mon + 1) >> 1),
                    (((UBYTE)t->tm_mon + 1) << 7) | (UBYTE)(t->tm_year % 100),
                    0,0) == FALSE)
        return FALSE ;
    /* Initialisation de la piste et de son image */
    if ((cc90.diskop == TRUE) && (dk[drive].drive <= 3))
        memset(track_buffer, 0xE5, TRACKSIZE*2);

    /* Rceptionne le bloc Sys du systme du Thomson */
    if (GetInfo() == FALSE) return FALSE ;

    cc90.config = cc90.info.data[0] ;
    cc90.lect   = cc90.info.data[1] ;
    cc90.ptrack = (((USHORT)cc90.info.data[3] << 8) | (USHORT)cc90.info.data[4]) ;

    i = 0 ;
    while ((cont[i].to != 0)
        && (cont[i].to != (cc90.lect & 0x07)))
        i++ ;
    dk[drive].cont = i ;

    switch(dk[drive].drive)
    {
        case 0 :
        case 1 :
        case 2 :
        case 3 :
            dk[drive].nbtracks = cont[i].nbtracks ;
            dk[drive].fatsize  = cont[i].fatsize  ;
            dk[drive].secsize  = (size_t)(((cc90.lect & 0x08) != 0) ? 128 : 256) ;
            break ;
        case 4 :
            if ((dk[drive].external == TRUE)
             || ((cc90.config & 0x70) == 0))
                return PError(dk[drive].arg,ERR_NO_DISK) ;
            dk[drive].nbtracks = cc90.config & 0x70  ;
            dk[drive].fatsize  = MAXFATSIZE  ;
            dk[drive].secsize  = MAXSECSIZE  ;
            break ;
        default :
            return PError(dk[drive].arg,ERR_NO_DISK) ;
    }

#ifdef DEBUG_MODE
    const struct {
        int   code ;
        char  name[7] ;
      } machine[7] = {
        { MO5_COMPUTER  , "MO5"    },
        { MO6_COMPUTER  , "MO6"    },
        { TO7_COMPUTER  , "TO7"    },
        { TO770_COMPUTER, "TO7-70" },
        { TO9_COMPUTER  , "TO9"    },
        { TO8_COMPUTER  , "TO8"    },
        { TO9P_COMPUTER , "TO9+"   }
    } ;
    const char densitylist[2][7] = { "Double" , "Simple" } ;
    fprintf(fp_log,"ToInit = cont[%d] config[$%02x] lect[$%02x] tracks[%d] fat[%d] sector[%d] ptrack[$%04x]\n",
                    dk[drive].cont,
                    (unsigned char)cc90.config,
                    (unsigned char)cc90.lect,
                    dk[drive].nbtracks,
                    dk[drive].fatsize,
                    dk[drive].secsize,
                    cc90.ptrack) ;
    fprintf(fp_log,"Config =\n") ;
    i = 0 ; while((i < 6) && ((cc90.config & 0x0F) != machine[i].code)) i++ ;
    fprintf(fp_log,"     Ordi = '%s'\n",machine[i].name) ;
    fprintf(fp_log,"     RAM  = %d kilooctets\n",((cc90.config >> 4) & 0x0F) * 64) ;
    fprintf(fp_log,"Lect =\n") ;
    fprintf(fp_log,"     Entrelacement = %d\n",(cc90.lect >> 4) & 0x0F) ;
    fprintf(fp_log,"     Densit = %s\n", densitylist[(cc90.lect >> 3) & 0x01]) ;
    
#endif

    return TRUE ;
}


/*
 * Rcupration d'une zone mmoire du Thomson
 */
static int GetMemory(UBYTE bank, UBYTE *datas, USHORT to_ptr, USHORT size)
{
    USHORT datasize ;
    USHORT pos = 0 ;

    while(pos < size)
    {
        /* Limite la taille des donnes */
        datasize = ((size - pos) > DATASIZE) ? DATASIZE : (size - pos) ;

        /* Demande l'envoi des donnes */
        if (SendInfo(CODE_GETMEMORY,
                     bank,
                     (UBYTE)((to_ptr + pos) >> 8),
                     (UBYTE)(to_ptr + pos),
                     (UBYTE)(datasize >> 8),
                     (UBYTE)datasize) == FALSE)
            return FALSE ;

        /* Rceptionne les donnes */
        if (GetDatas(datas + pos, datasize) == FALSE) return FALSE ;

        pos += datasize ;
    }
    return TRUE ;
}


/*
 * Dfinit le numro de banque et le pointeur du premier secteur RamDisk
 */
static int RamDiskPointers(int drive, UBYTE track, UBYTE first, UBYTE *bank, USHORT *to_ptr)
{
    UBYTE tracknum ;
    UBYTE bankmax ;

    /* Erreur si pas de RamDisk */
    if ((cc90.config & 0xF0) == 0)
        return PError(dk[drive].arg,ERR_IO_ERROR) ;

    switch(cc90.config & 0x0F)
    {
        case TO9P_COMPUTER :
        case TO8_COMPUTER  :
            /* Conversion du numro de piste */
            tracknum = (track == 20) ? 0 : (track + ((track < 20) ? 1 : 0)) ;

            /* Actualise le numro de banque */
            *bank = tracknum >> 2 ;

            /* Limite l'accs  l'intervalle de banque */
            bankmax  = ((cc90.config & 0x80) >> 3) | 0x0F ;
            *bank = bankmax - (tracknum >> 2) ;
            if ((*bank < (bankmax - (cc90.config & 0x70)))
             || (*bank > bankmax))
                return PError(dk[drive].arg,ERR_IO_ERROR) ;

            /* Calcule le pointeur de secteur
             * Comme le RamDisk est trait en espace
             * cartouche et que l'on doit faire avec
             * les espaces RAM, le bit 13 du pointeur
             * est commut, les moitis de RAM tant
             * inverses */
            *to_ptr = (USHORT)(((((tracknum & 0x03) << 12) + ((first - 1) * MAXSECSIZE)) ^ 0x2000) + 0xA000) ;
            break ;

        case TO9_COMPUTER :
            /* Limite l'accs  l'intervalle de banque */
            if ((track < 16) || (track > 31))
                return PError(dk[drive].arg,ERR_IO_ERROR) ;

            /* Actualise le numro de banque */
            *bank = ((track - 16) >> 2) + 8 ;

            /* Calcule le pointeur de secteur
             * Comme le RamDisk est trait en espace
             * RAM sur le TO9, nul n'est donc besoin
             * de commuter le bit 13 du pointeur */
            *to_ptr = (USHORT)((((track & 0x03) << 12) + ((first - 1) * MAXSECSIZE)) + 0xA000) ;
            break ;

        default :
            return PError(dk[drive].arg,ERR_IO_ERROR) ;
    }
    return TRUE ;
}


/*
 * Transfre une srie de secteurs Thomson
 */
int /*@unused@*/ ToGetSectors(int drive, UBYTE track, UBYTE first, UBYTE last)
{
    UBYTE /*@unsignedintegraltype@*/ bank = 0 ;
    USHORT /*@unsignedintegraltype@*/ ptr = 0 ;

    switch (dk[drive].drive)
    {
        case 4 :
            /* Pour le RamDisk, les secteurs sont prlevs
             * directement dans la piste */
            if (RamDiskPointers(drive, track, first, &bank, &ptr) == FALSE) return FALSE ;
            return GetMemory(bank - 1,
                             &track_buffer[(first - 1) * MAXSECSIZE],
                             ptr,
                             (USHORT)((last - first + 1) * MAXSECSIZE)) ;
        default :
            /* Rcupration des secteurs de piste */
            if ((GetMemory(CURRENT_BANK,
                          &track_buffer[(first - 1) * dk[drive].secsize],
                          cc90.ptrack + ((first - 1) * dk[drive].secsize),
                          (last - first + 1) * dk[drive].secsize)) == FALSE)
                return FALSE ;
            /* Mise  jour de l'image de piste */
            memcpy(&track_buffer[TRACKSIZE + (first - 1) * dk[drive].secsize],
                   &track_buffer[(first - 1) * dk[drive].secsize],
                   (last - first + 1) * dk[drive].secsize) ;
            break ;
    }
    return TRUE ;
}

/*
 * Lit une srie de secteurs Thomson
 */
static int ToReadSectors(int drive, UBYTE track, UBYTE first, UBYTE last)
{
    /* Chargement des secteurs de piste si floppy */
    if (dk[drive].drive <= 3)
        if (SendInfo(CODE_LOADSECTORS, dk[drive].drive, track,  first, last, 0) == FALSE)
            return FALSE ;

    return ToGetSectors(drive, track, first, last) ;
}


/*
 * Envoi d'une zone mmoire du Thomson
 */
int /*@unused@*/ SendMemory(UBYTE bank, UBYTE *datas, USHORT to_ptr, USHORT size)
{
    USHORT datasize ;
    USHORT pos = 0 ;

    while(pos < size)
    {
        /* Limite la taille des donnes */
        datasize = ((size - pos) > DATASIZE) ? DATASIZE : (size - pos) ;

        /* Envoie le bloc d'info */
        if (SendInfo(CODE_SENDMEMORY,
                     bank,
                     (UBYTE)((to_ptr + pos) >> 8),
                     (UBYTE)to_ptr + pos,
                     (UBYTE)(datasize >> 8),
                     (UBYTE)datasize) == FALSE)
            return FALSE ;

        /* Envoie les donnes */
        if (SendDatas(datas + pos, datasize) == FALSE) return FALSE ;

        pos += datasize ;
    }
    return TRUE ;
}


/*
 * Sauvegarde d'une srie de secteurs au format Thomson (srie)
 */
int /*@unused@*/ ToPutSectors(int drive, UBYTE track, UBYTE first_sector, UBYTE last_sector)
{
    USHORT i ;
    int rep ;
    UBYTE first = 0, last = 0 ;
    USHORT pos = (USHORT)(first_sector - 1) * dk[drive].secsize ;
    UBYTE /*@unsignedintegraltype@*/ bank = 0 ;
    USHORT /*@unsignedintegraltype@*/ ptr = 0 ;

    switch(dk[drive].drive)
    {
        case 4 :
            /* Pour le RamDisk, les secteurs envoys
             * sont dposs directement dans la piste */
            if (RamDiskPointers(drive, track, first_sector, &bank, &ptr) == FALSE) return FALSE ;
            return (SendMemory(bank - 1,
                               &track_buffer[(first_sector - 1) * MAXSECSIZE],
                               ptr,
                               (last_sector - first_sector + 1) * MAXSECSIZE)) ;
        default :
            /* Optimisation du transfert pour floppy
             * Le programme vrifie si les secteurs sont identiques
             *  ceux qui se trouvent en lieu et place dans la mmoire
             * de piste du Thomson (il garde une image de la piste)
             * Il n'enverra au Thomson que les secteurs diffrents */
            for(i=first_sector;i<=last_sector;i++)
            {
                /* Reprage des secteurs diffrents */
                if ((rep = memcmp(&track_buffer[pos],&track_buffer[TRACKSIZE + pos],dk[drive].secsize)) != 0)
                {
                    if (first == 0) first = i ;
                    last = i ;
                }
                /* Enregistrement ventuel des secteurs */
                if ((rep == 0) || (i == last_sector))
                {
                    if (first != 0)
                        /* Envoi des secteurs de piste */
                        if (SendMemory(CURRENT_BANK,
                                       &track_buffer[(first - 1) * dk[drive].secsize],
                                       cc90.ptrack + ((first - 1) * dk[drive].secsize),
                                       (last - first + 1) * dk[drive].secsize) == FALSE)
                            return FALSE ;
                    first = 0 ;
                }
                /* Actualisation de l'image de piste */
                memcpy(&track_buffer[TRACKSIZE + pos], &track_buffer[pos], dk[drive].secsize) ;
                /* Secteur suivant */
                pos += dk[drive].secsize ;
            }
            break ;
    }
    return TRUE ;
}

/*
 * Sauvegarde d'une srie de secteurs au format Thomson (srie)
 */
static int ToSaveSectors(int drive, UBYTE track, UBYTE first, UBYTE last)
{
    if (ToPutSectors(drive,track,first,last) == FALSE)
        return FALSE ;

    /* Sauvegarde de la piste par le Thomson */
    if (dk[drive].drive < 4)
        return SendInfo(CODE_SAVESECTORS, dk[drive].drive, track, first, last, 0) ;
    return TRUE ;
}


/* ========================================================================== *
 *                            OPERATIONS SAP
 * ========================================================================== */

/*
 * Initialisation du drive SAP
 */
static int
SapInit(int drive) /*@globals errno@*/
{
    char string[strlen(SapString)+1] ;
    int i ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"SapInit = ") ;
#endif

    dk[drive].nbtracks = 0 ;

    /* Ouvre le fichier SAP */
    if (CreateIfNotExist(drive) == FALSE) return FALSE ;

    /* Vrifie la validit du SAP */
    if (OpenFpFile(FALSE, dk[drive].dkname,"rb") == FALSE) return FALSE ;
    string[strlen(SapString)] = '\0' ;
    (void)fread(string, 1, strlen(SapString), fp_file) ;
    CloseFpFile() ;

    if(strcmp(&string[1],&SapString[1]) != 0)
        return PError(dk[drive].dkname,ERR_BAD_FILE_FORMAT) ;

    /* Rcupre les spcifications */
    dk[drive].nbtracks = 0 ;
    i = 0 ;
    while ((cont[i].to != 0) && (cont[i].sap != (UBYTE)string[0]))
        i++ ;
    dk[drive].cont     = i ;
    dk[drive].nbtracks = cont[i].nbtracks ;
    dk[drive].fatsize  = cont[i].fatsize  ;
    dk[drive].secsize  = cont[i].secsize  ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"cont[%d] tracks[%d] fat[%d] sector[%d]\n",
                    dk[drive].cont,
                    dk[drive].nbtracks,
                    dk[drive].fatsize,
                    dk[drive].secsize) ;
#endif

    if ((dk[drive].nbtracks == 0) || (dk[drive].drive > 0))
        return PError(dk[drive].arg,ERR_NO_DISK) ;

    return TRUE ;
}


/*
 * Oprations de lecture ou d'criture de secteurs SAP
 */
static int
SapOp(int mode, int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
    long pos = (long)((track * SECTORS_PER_TRACK + first - 1)
                    * (dk[drive].secsize + 6)
                    + strlen(SapString)) ;
    UBYTE sector ;
    UBYTE buf[MAXSECSIZE + 6] ;

    if(track >= dk[drive].nbtracks)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;

    for(sector=first;sector<=last;sector++)
    {
        /* Charge le secteur SAP */
        if (OpenFpFile(FALSE, dk[drive].dkname,"rb") == FALSE) return FALSE ;
        if (fseek(fp_file, pos, SEEK_SET) != 0)
            return PError(dk[drive].dkname,ERR_IO_ERROR) ;
        if (fread(buf, 1, dk[drive].secsize+6, fp_file) != dk[drive].secsize+6)
            return PError(dk[drive].dkname,ERR_IO_ERROR) ;
        CloseFpFile() ;

        /* Vrifie la validit des infos secteur */
        if ((buf[2] != track) || (buf[3] != sector))
            return PError(dk[drive].dkname,ERR_IO_ERROR) ;

        /* Dcode le secteur et vrifie le CRC */
        DoMagic(&buf[4], dk[drive].secsize) ;
        GetCrcPuk(buf, dk[drive].secsize+4) ;
        if(((buf[dk[drive].secsize+4] << 8)
          | buf[dk[drive].secsize+5]) != crcpuk_temp)
            return PError(dk[drive].dkname,ERR_CRC_ERROR) ;

        /* Oprations si lecture ou criture */
        switch(mode)
        {
            case OP_READ :
                /* Copie le secteur sur la piste */
                memcpy(&track_buffer[(sector - 1) * dk[drive].secsize],
                       &buf[4],
                       dk[drive].secsize) ;
                break ;
            case OP_WRITE :
                /* Cre le secteur SAP */
                BuildSapSector(sector,buf,dk[drive].secsize) ;

                /* Sauve le secteur SAP */
                if (OpenFpFile(FALSE, dk[drive].dkname,"rb+") == FALSE) return FALSE ;
                if (fseek(fp_file, pos, SEEK_SET) != 0)
                    return PError(dk[drive].dkname,ERR_IO_ERROR) ;
                if (fwrite(buf, 1, dk[drive].secsize+6, fp_file) != dk[drive].secsize+6)
                    return PError(dk[drive].dkname,ERR_IO_ERROR) ;
                CloseFpFile() ;
                break ;
        }

        pos += dk[drive].secsize+6 ;
    }
    return TRUE ;
}


/*
 * Lit les secteurs SAP
 */
static int
SapReadSectors(int drive, UBYTE track, UBYTE first, UBYTE last)  /*@globals errno@*/
{
    return SapOp(OP_READ, drive, track, first, last) ;
}


/*
 * Ecrit les secteurs SAP
 */
static int
SapSaveSectors(int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
    return SapOp(OP_WRITE, drive, track, first, last) ;
}


/* ========================================================================== *
 *                            OPERATIONS DCMOTO
 * ========================================================================== */


/*
 * Initialisation du drive DCMOTO
 */
static int
DcInit(int drive) /*@globals errno@*/
{
    long filesize;
    int i ;
    UBYTE maxdrive = 0 ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"DcInit = ") ;
#endif

    /* Ouvre le fichier DCMOTO */
    if (CreateIfNotExist(drive) == FALSE) return FALSE ;

    /* Vrifie la validit du fichier DCMOTO */
    if (OpenFpFile(FALSE, dk[drive].dkname,"rb") == FALSE) return FALSE ;
    if (fseek(fp_file, 0, SEEK_END) != 0)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;
    if ((filesize = ftell(fp_file)) < 0)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;
    CloseFpFile() ;

    /* Rcupre les spcifications */
    for(i=0;i<CONTSIZE;i++)
    {
        if(filesize == cont[i].dcmoto)
        {
            maxdrive           = cont[i].maxdrive  ;
            dk[drive].cont     = i ;
            dk[drive].nbtracks = cont[i].nbtracks ;
            dk[drive].fatsize  = cont[i].fatsize  ;
            dk[drive].secsize  = cont[i].secsize  ;
        }
    }

#ifdef DEBUG_MODE
    fprintf(fp_log,"cont[%d] disks[%d] tracks[%d] fat[%d] sector[%d]\n",
                    dk[drive].cont,
                    maxdrive,
                    dk[drive].nbtracks,
                    dk[drive].fatsize,
                    dk[drive].secsize) ;
#endif

    if ((dk[drive].nbtracks == 0) || (dk[drive].drive > maxdrive))
        return PError(dk[drive].arg,ERR_IO_ERROR) ;

    return TRUE ;
}


/*
 * Oprations de lecture ou d'criture de secteurs DCMOTO
 */
static int
DcOp(int mode, int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
    long pos;
    char filemode[5] = "rb\0" ;
    size_t size = dk[drive].secsize*(last-first+1) ;

    if (track >= dk[drive].nbtracks)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;

    if (mode == OP_WRITE) strcat(filemode,"+") ;

    if (OpenFpFile(FALSE, dk[drive].dkname,filemode) == FALSE)
        return FALSE ;

    pos = (long)(((dk[drive].drive * dk[drive].nbtracks + track)
                 * SECTORS_PER_TRACK)
                 + (first - 1)) ;

    if (fseek(fp_file, pos*dk[drive].secsize, SEEK_SET) != 0)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;

    switch(mode)
    {
        /* Copie le secteur sur la piste */
        case OP_READ :
             if (fread(&track_buffer[(first-1)*dk[drive].secsize], 1, size, fp_file) != size)
                 return PError(dk[drive].dkname,ERR_IO_ERROR) ;
             break ;
        /* Sauve le secteur SAP */
        case OP_WRITE :
             if (fwrite(&track_buffer[(first-1)*dk[drive].secsize], 1, size, fp_file) != size)
                 return PError(dk[drive].dkname,ERR_IO_ERROR) ;
             break ;
    }

    CloseFpFile() ;

    return TRUE ;
}


/*
 * Lit les secteurs DCMOTO
 */
static int
DcReadSectors(int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
    return DcOp(OP_READ, drive, track, first, last) ;
}


/*
 * Ecrit les secteurs DCMOTO
 */
static int
DcSaveSectors(int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
    return DcOp(OP_WRITE, drive, track, first, last) ;
}


/* ========================================================================== *
 *                            FONTIONS PUBLIQUES
 * ========================================================================== */


/*
 * Initialisation de tous les lecteurs
 */
int
InitDrives(void) /*@globals errno@*/
{
    int drive ;

    /* Initialisation des lecteurs */
    for(drive = 0; drive < 4; drive++)
    {
        if ((dk[drive].type != INVALID_TYPE)
         && (dk[drive].type != FILE_TYPE))
        {
            switch(dk[drive].type)
            {
                case TO_TYPE :
                    if (ToInit(drive) == FALSE) return FALSE ;
                    break ;
                case SAP_TYPE :
                    if (SapInit(drive) == FALSE) return FALSE ;
                    break ;
                case DCMOTO_TYPE :
                    if (DcInit(drive) == FALSE) return FALSE ;
                    break ;
            }
            if (dos_buffer == NULL) 
                return PError("",ERR_NOT_ENOUGH_MEMORY) ;
            dk[drive].fat = dos_buffer + (drive * TRACKSIZE) + ((FATSECTOR - 1) * dk[drive].secsize) ;
            memset(dos_buffer + (drive * TRACKSIZE),0x00,TRACKSIZE) ;
        }
    }
    return TRUE ;
}


/*
 * Chargement d'une srie de secteurs
 */
int
ReadSectors(int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
#ifdef DEBUG_MODE
    fprintf(fp_log,"ReadSectors %d: (%d [%d-%d])\n", drive, track, first, last) ;
#endif

    switch(dk[drive].type)
    {
        case TO_TYPE :
            return ToReadSectors(drive, track, first, last) ;
        case SAP_TYPE :
            return SapReadSectors(drive, track, first, last) ;
        case DCMOTO_TYPE :
            return DcReadSectors(drive, track, first, last) ;
    }
    return PError(dk[drive].dkname,ERR_BAD_FILE_FORMAT) ;
}


/*
 * Sauvegarde d'une srie de secteurs
 */
int
SaveSectors(int drive, UBYTE track, UBYTE first, UBYTE last) /*@globals errno@*/
{
#ifdef DEBUG_MODE
    fprintf(fp_log,"SaveSectors %d: (%d [%d-%d])\n", drive, track, first, last) ;
#endif
    switch(dk[drive].type)
    {
        case TO_TYPE :
            return ToSaveSectors(drive, track, first, last) ;
        case SAP_TYPE :
            return SapSaveSectors(drive, track, first, last) ;
        case DCMOTO_TYPE :
            return DcSaveSectors(drive, track, first, last) ;
    }
    return PError(dk[drive].arg,ERR_BAD_FILE_FORMAT) ;
}

