/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Parser trs simplifi pour lire les fichiers INI
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "defs.h"

#ifdef _GNU_SOURCE
#   define INIPATH "/etc/cc90/cc90.ini"
#else
#   define INIPATH "c:\\windows\\cc90\\cc90.ini"
#endif

static char linebuffer[300] ; /* Buffer de ligne */
static char *line ;           /* Pointeur de travail sur linebuffer */
static UBYTE *inibuf ;         /* Pointeur de travail sur file_buffer */

extern struct CC90_T cc90 ;
extern struct OPT_T opt ;
extern struct CONV_T conv ;

extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern void FreeFile(void) ;

extern int PError(char *, int) ;
extern int LoadFpFile(char *) ;
extern int InitConversionFormat(char *) ;

/*
 * Capture d'une ligne du fichier
 */
static int GetIniLine(void)
{
   line = linebuffer ;

   if (inibuf >= (file_buffer + file_size)) return FALSE ;

   while ((inibuf < (file_buffer + file_size))
       && (line < (linebuffer + STRING_SIZE_MAX - 1))
       && (*inibuf != 0x0A)
       && (*inibuf != 0x0D)
       && (*inibuf != 0x00))
       *(line++) = (char)*(inibuf++) ;

   while ((inibuf < (file_buffer + file_size))
       && (*inibuf != 0x0A)
       && (*inibuf != 0x0D)
       && (*inibuf != 0x00))
       inibuf++ ;

   /* Clture la ligne */
   *line = '\0' ;

   switch(*inibuf)
   {
       case 0x0A : if (*(inibuf+1) == 0x0D) inibuf++ ; break ;
       case 0x0D : if (*(inibuf+1) == 0x0A) inibuf++ ; break ;
   }
   inibuf++ ;
   line = linebuffer ;

   return TRUE ;
}


/*
 * Recherche la section
 */
static int FindIniSection(char *name)
{
    char section_name[strlen(name)+3] ;

    (void)snprintf(section_name,strlen(name)+3,"[%s]",name) ;

    /* Recherche le nom de la section */
    inibuf = file_buffer ;
    while (GetIniLine() == TRUE)
        if (strncmp(section_name,line,strlen(section_name)) == 0)
            return TRUE ;
    return PError(name,ERR_SECTION_NOT_FOUND) ;
}


/*
 * Recherche la clef
 */
static int FindIniKey(char *name)
{
    /* Recherche le nom de la clef */
    while (GetIniLine() == TRUE)
    {
        if (*inibuf == (UBYTE)'[')
            return PError(name,ERR_KEY_NOT_FOUND) ;
        else
        if (strncmp(name,line,strlen(name)) == 0)
        {
            line += strlen(name) ;
            if (*(line++) != '\x20')
                return PError(name,ERR_SPACE_AFTER_KEY) ;
            if (*(line++) != '=')
                return PError(name,ERR_EQUAL_AFTER_KEY) ;
            while((*line == '\x20')
               || (*line == '\x09'))
                line++ ;
            return TRUE ;
        }
    }
    return PError(name,ERR_KEY_NOT_FOUND) ;
}


/*
 * Recherche la chane de la valeur
 */
static int GetString(char *section_name, char *key_name, char *ptvalue, int maxsize)
{
    int i = 0 ;

    if (FindIniSection(section_name) == FALSE)
        return FALSE ;
    if (FindIniKey(key_name) == FALSE)
        return FALSE ;

    while((i < maxsize)
       && (*line != ';')
       && (*line > '\x20'))
        ptvalue[i++] = *(line++) ;
    ptvalue[i] = '\0' ;

    return TRUE ;
}


/*
 * Rcupre une valeur dcimale dans un INT
 */
static int GetInt(char *section_name, char *key_name, int *ptvalue)
{
    int i ;
    char string[7] = "" ;
    if (GetString(section_name,key_name,string,6) == FALSE)
        return FALSE ;
    for(i=0;i<(int)strlen(string);i++)
        if (isdigit((int)string[i]) == 0)
            return PError(string,ERR_VALUE_NOT_DECIMAL) ;
    if (string[0] != '\0')
        *ptvalue = atoi(string) ;
    return TRUE ;
}


static void UpperString(char *string)
{
    int i ;
    for(i=0;i<(int)strlen(string);i++)
        string[i]=(char)toupper((int)string[i]) ;
}


static int GetYesNo(char *section_name, char *key_name, int *ptvalue)
{
    char string[52] = "" ;
    if (GetString(section_name,key_name,string,50) == FALSE)
        return FALSE ;
    UpperString(string) ;
    *ptvalue = (strcmp(string,"YES") == 0) ? TRUE : FALSE ;
    return TRUE ;
}

static int GetConversion(char *section_name, char *key_name)
{
    char string[52] = "" ;
    if (GetString(section_name,key_name,string,50) == FALSE)
        return FALSE ;
    if (InitConversionFormat(string) == FALSE)
        return PError(string,ERR_UNKNOWN_VALUE) ;
    return TRUE ;
}

/*
 * Charge et traite le fichier INI
 */
int LoadIni(void)
{
    int success = FALSE ;
    if (LoadFpFile(INIPATH) == FALSE)
        return FALSE ;
    if ((GetString("Cc90","port.name",cc90.portname,39) == TRUE)  /* Nom du port srie */
     && (GetInt("Cc90","timeout.max",&cc90.timeout_max) == TRUE)  /* Timeout maximum */
     && (GetInt("Cc90","timeout.mid",&cc90.timeout_mid) == TRUE)  /* Timeout mdian */
     && (GetInt("Disk","interlace",&opt.interlace) == TRUE)  /* Entrelacement par dfaut */
     && (GetConversion("Files","default") == TRUE)
     && (GetYesNo("Files","gamma.correction",&conv.gammac) == TRUE)) /* Conversions par dfaut */
          success = TRUE ;
    FreeFile() ;
    return success ;
}

