/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Primitives srie pour systmes Unix
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sysexits.h>
#include <sys/time.h>

#ifndef TRUE
#   define TRUE 1
#endif

#ifndef FALSE
#   define FALSE 0
#endif

static struct termios gOriginalTTYAttrs;
static int fileDescriptor = -1 ;
static long timeout = 50 ;

/*
 * Ferme le port srie
 */
void
SerialClose(void) /*@globals errno@*/
{
    if (fileDescriptor != -1)
    {
        /* Flushe le buffer */
        (void)tcdrain(fileDescriptor) ;

        /* Rtablit l'ancien port */
        (void)tcsetattr(fileDescriptor, TCSANOW, &gOriginalTTYAttrs) ;

        /* Ferme le port */
        (void)close(fileDescriptor);
    }
    fileDescriptor = -1 ;
}


/*
 * Ouvre le port srie
 */
int
SerialOpen(char *portname, long timelimit) /*@globals errno@*/
{
    struct termios  options;

    /* Sortie si dj ouvert */
    if (fileDescriptor != -1)
        return TRUE;
    timeout = timelimit ;

    /* Ouverture du port srie */
    if ((fileDescriptor = open(portname, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0)
        return FALSE;

    /* Rservation du port srie pour un usage exclusif */
    if (ioctl(fileDescriptor, TIOCEXCL) < 0)
        return FALSE;

    /* Demande le bloquage lors de l'usage */
    if (fcntl(fileDescriptor, F_SETFL,  NULL) < 0)
        return FALSE;

    /* Sauvegarde options courantes pour restitution */
    if (tcgetattr(fileDescriptor, &gOriginalTTYAttrs) < 0)
        return FALSE;

    /* Programmation du port srie */
    options = gOriginalTTYAttrs;
    cfmakeraw(&options);
    options.c_cc[VTIME] = timeout/100 ;  /* timeout (read) en diximes de secondes */
    options.c_cc[VMIN] = 0 ;    /* Sort si timeout */
    (void)cfsetispeed(&options,B38400) ;
    (void)cfsetospeed(&options,B38400) ;
    options.c_cflag &= ~(PARENB | CSTOPB) ;
    options.c_cflag |= (CSIZE | CS8 | CRTSCTS | CREAD | CLOCAL) ;
    options.c_iflag &= ~(IXON | IXOFF | IXANY) ;
    options.c_iflag |= (IGNBRK | IGNPAR) ;
    options.c_oflag &= ~(OPOST) ;
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG) ;

    /* Activation du port srie */
    tcflush(fileDescriptor, TCIFLUSH);
    if (tcsetattr(fileDescriptor, TCSANOW, &options) < 0)
        return FALSE;

    /* Attente de suret */
    usleep(600) ;

    return TRUE ;
}


/*
 * Envoie le bloc de donnes asynchrones sur le port srie
 *
 * Le buffer de donnes est crit octet par
 * octet pour prvenir l'overflow du
 * buffer de l'adaptateur USB-Srie
 */
int
SerialWrite(char *buf, int size) /*@globals errno@*/
{
    register int i;

    /* Le retour de 'write' ne doit pas tre :
     * <0 : une erreur d'criture est survenue
     * =0 : aucune donne n'a t crite */
    for(i=0;i<size;i++)
    {
        /* Envoie la donne au Thomson */
        if (write(fileDescriptor, buf+i, 1) <= 0)
            return FALSE ;
    }
    return TRUE ;
}


/*
 * Reoit un bloc de donnes asynchrones du port srie
 *
 * Le buffer de donnes est lu octet par
 * octet pour prvenir l'overflow du
 * buffer de l'adaptateur USB-Srie
 */
int
SerialRead(char *buf, int size) /*@globals errno@*/
{
    register int i ;
    /* Le retour de 'read' ne doit pas tre :
     * <0 : une erreur de lecture est survenue
     * =0 : le timeout est dpass */
    for(i=0;i<size;i++)
    {
       /* Reoie la donne du Thomson */
        if (read(fileDescriptor, buf+i, 1) <= 0)
            return FALSE ;
    }
    return TRUE ;
}
