/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Programme principal
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include <dirent.h>

#include "defs.h"

#ifdef DEBUG_MODE
    FILE *fp_log ;
    FILE *fp_map ;
#endif
FILE *fp_file ;

struct CC90_T cc90 ;
struct DISK_T dk[4] ;
struct OPT_T opt ;
struct CONV_T conv ;

extern const struct CONTROLLER cont[] ;

/* Fonctions du programme */
extern void SerialInstall(void) ;
extern void CopyDisk(void) ;
extern void CopyFile(void) ;
extern void SendBinary(void) ;
extern void FormatDisk(void) ;
extern void DeleteFiles(void) ;
extern void DrawDir(void) ;
extern void FormatDisk(void) ;
extern void CreateDisk(void) ;

extern int InitDrives(void) ;
extern int Chop(char *) ;
extern void SerialClose(void) ;

extern int OpenFpFile(int, char *, char *) ;
extern void CloseFpFile(void) ;

extern int LoadIni(void) ;

extern int AllocTracks(void) ;
extern void FreeAll(void) ;

/*
 * Ferme toutes les dclarations
 */
int /*@unused@*/ CloseAll(void)
{
    CloseFpFile() ;
    SerialClose() ;
    FreeAll() ;
#ifdef DEBUG_MODE
    if (fp_log) fclose(fp_log) ;
    fp_log = NULL ;
    if (fp_map) fclose(fp_map) ;
    fp_map = NULL ;
#endif
    return 0 ;
}


/*
 * Initialisation des structures
 */
static void InitStructures(void)
{
    int drive ;

    /* Init structure 'cc90' */
    cc90.exit = FALSE ;
    cc90.yes = FALSE ;

    /* Init structure 'opt' */
    opt.autochg   = 0x00 ;
    opt.interlace = 7 ;
    opt.quick = FALSE ;

    /* Init structure 'dk' */
    for(drive=0;drive<4;drive++)
    {
        dk[drive].type = INVALID_TYPE ;
        dk[drive].cont = 4 ;
        dk[drive].drive = 0 ;
        dk[drive].external = FALSE ;
    }

    /* Init structure 'conv' */
    conv.maptype = CONV_GFX_SNP ;
    conv.txttype = CONV_TXT_ASM ;
    conv.sndtype = CONV_SND_SND ;
    conv.gammac = TRUE ;
    conv.sample_rate = 8000 ;
    conv.force40cols = FALSE ;
    
    /* Chargement de la configuration par dfaut */
    if (LoadIni() == FALSE) exit(CloseAll()) ;
}


/*
 * Affiche l'en-tte de l'aide
 */
static void PrintHelpHead(void)
{
    (void)printf("\n") ;
    (void)printf("Transferts serie pour Thomson\n") ;
    (void)printf("  Prehisto & Oncletom (c) CC90 v %s.%s%s - %s\n",
                        CC90_VERSION0,
                        CC90_VERSION1,
                        CC90_VERSION2,
                        CC90_DATE) ;
}


/*
 * Affichage de l'aide
 */
static void PrintHelp(void)
{
    PrintHelpHead() ;
    (void)printf("\n") ;
    (void)printf("CC90 --commande [-options] [-conversion] source [dest [source dest]]\n") ;
    (void)printf("\n") ;
    (void)printf("  --commande :\n") ;
    (void)printf("    --help    aide du programme\n") ;
    (void)printf("    --install installation sans disquette\n") ;
    (void)printf("    --create  creation d'une disquette virtuelle\n") ;
    (void)printf("    --format  formatage de disquette\n") ;
    (void)printf("    --delete  effacement de fichiers Thomson\n") ;
    (void)printf("    --dir     affichage du repertoire\n") ;
    (void)printf("    --disk    copie de disquette\n") ;
    (void)printf("    --file    copie de fichiers\n") ;
    (void)printf("    --mem     implantation memoire\n") ;
    (void)printf("\n") ;
    (void)printf("  -options :\n") ;
    (void)printf("    -d    + type de disquette a creer (0-7)\n") ;
    (void)printf("    -a    lancement automatique pour fichiers CHG\n") ;
    (void)printf("    -b    + numero de banque pour implantation memoire (0-31)\n") ;
    (void)printf("    -e    + entrelacement formatage (1-15)\n") ;
    (void)printf("    -q    formatage/copie logique de disquette\n") ;
    (void)printf("    -c    (+) conversion fichier (ASM/ASC/SND/SON/MAP/SND/PPM)\n") ;
    (void)printf("  -conversion :\n") ;
    (void)printf("    -s    + debit pour le fichier WAV\n") ;
    (void)printf("    -g    Switch de gamma correction pour les graphiques\n") ;
    (void)printf("    -0    + force graphique en 40 colonnes si possible\n") ;
    exit(CloseAll()) ;
}


/*
 * Affiche l'en-tte du conseil
 */
static void PrintAdviceHead(void)
{
    (void)printf("Taper 'cc90 --help' pour afficher l'aide\n") ;
}


/*
 * Affichage du conseil
 */
static void PrintAdvice(char *arg, char *errorstring)
{
    (void)printf("*** %s : %s\n", arg, errorstring) ;
#ifdef DEBUG_MODE
    fprintf(fp_log, "*** %s : %s\n", arg, errorstring) ;
#endif
    PrintAdviceHead() ;
    exit(CloseAll()) ;
}


/*
 * Liste des erreurs
 */
static const struct ERRORMESSAGE {
    int     number ;
    char    string[33] ;
  } Error_T[] = {
    { ERR_PORT_NOT_READY,     "Port serie non pret"              },
    { ERR_BREAK,              "Operation interrompue"            },
    { ERR_SERIAL_ERROR,       "Erreur de transmission serie"     },
    { ERR_NOT_ENOUGH_MEMORY,  "Pas assez de memoire"             },
    { ERR_CRC_ERROR,          "Erreur de CRC"                    },
    { ERR_SAME_DISKS,         "Source et destination identiques" },
    { ERR_ILLEGAL_WILDCARDS,  "Jokers non autorises"             },
    { ERR_FATFULL,            "Place sur disquette insuffisante" },
    { ERR_FILE_NOT_FOUND,     "Fichier(s) introuvable(s)"        },
    { ERR_DIR_NOT_FOUND,      "Repertoire introuvable"           },
    { ERR_DIRECTORY_FULL,     "Repertoire plein"                 },
    { ERR_BAD_FILE_NAME,      "Nom de fichier errone"            },
    { ERR_EMPTY_FILE_NAME,    "Le nom de fichier est vide"       },
    { ERR_BAD_FILE_FORMAT,    "Format de fichier non reconnu"    },
    { ERR_CANT_OPEN_FILE,     "Impossible d'ouvrir le fichier"   },
    { ERR_BAD_FAT,            "FAT incoherente"                  },
    { ERR_BAD_ENTRY,          "Entree de fichier incoherente"    },
    { ERR_NO_DISK,            "Le disque n'existe pas"           },
    { ERR_IO_ERROR,           "Erreur d'entree/sortie"           },
    { ERR_INCOMPATIBLE_DISK,  "Disque incompatible"              },
    { ERR_ILLEGAL_OPERATION,  "Operation interdite"              },
    { ERR_COPY_DIR_INTO_FILE, "Copie de repertoire sur fichier"  },
    { ERR_PERMISSION_DENIED,  "Erreur de permission"             },
    { ERR_FILE_IS_A_DIR,      "Le fichier est un repertoire"     },
    { ERR_KEY_NOT_FOUND,      "Clef INI introuvable"             },
    { ERR_SPACE_AFTER_KEY,    "Espace manquant apres la clef"    },
    { ERR_EQUAL_AFTER_KEY,    "Signe '=' manquant"               },
    { ERR_SECTION_NOT_FOUND,  "Section INI introuvable"          },
    { ERR_VALUE_NOT_DECIMAL,  "Valeur INI non decimale"          },
    { ERR_UNKNOWN_VALUE,      "Valeur INI inconnue"              },
    { ERR_BIN_FILE_TOO_LONG,  "Fichier BIN trop long"            },
    { ERR_RESOLUTION,         "Erreur de resolution"             },
    { ERR_COMPRESSION,        "Erreur de compresssion"           }
};
#define ERRORTSIZE (int)sizeof(Error_T)/(int)sizeof(Error_T[0])


/*
 * Affichage de l'erreur
 */
int /*@unused@*/ PError(char *string, int errnum)
{
    int i = 0 ;
    char message_string[STRING_SIZE_MAX] = "*** " ;
    char result_string[STRING_SIZE_MAX] = "" ;

    /* Mmorise l'erreur */
    cc90.lasterror = errnum ;

    /* Enregistre le message joint */
    if(string[0] != '\0')
        (void)snprintf(message_string,STRING_SIZE_MAX-1,"*** '%s' : ",string) ;

    /* Enregistre l'erreur */
    while((errnum != Error_T[i].number) && (i<ERRORTSIZE))
        i++ ;
    if(i != ERRORTSIZE)
        (void)snprintf(result_string,STRING_SIZE_MAX-1,"%s%s",
                               message_string,
                               (char *)Error_T[i].string) ;
    else
        (void)snprintf(result_string,STRING_SIZE_MAX-1,"%sErreur inconnue",
                               message_string) ;

    /* Affiche l'erreur */
    (void)printf("%s\n",result_string) ;
#ifdef DEBUG_MODE
    fprintf(fp_log,"%s\n",result_string) ;
#endif
    return FALSE ;
}


enum {
    HELP_CHOICE,
    INSTALL_CHOICE,
    CREATE_CHOICE,
    FORMAT_CHOICE,
    DELETE_CHOICE,
    DISK_CHOICE,
    FILE_CHOICE,
    MEM_CHOICE,
    DIR_CHOICE
} ;
static const struct {
    char   command[12] ;
    int    maxnbfiles ;
    void   (*prog)(void) ;
    int    mseconds ;
    int    initdrives ;
    UBYTE  initthomson ;
  } choice_t[] = {
    { "--help"   , 0, PrintHelp    , MIN_WAIT_TIME, FALSE, FALSE },
    { "--install", 0, SerialInstall, MIN_WAIT_TIME, FALSE, FALSE },
    { "--create" , 1, CreateDisk   , MIN_WAIT_TIME, FALSE, FALSE },
    { "--format" , 1, FormatDisk   , MIN_WAIT_TIME, TRUE , FALSE },
    { "--delete" , 1, DeleteFiles  , MID_WAIT_TIME, TRUE , TRUE  },
    { "--disk"   , 4, CopyDisk     , MAX_WAIT_TIME, TRUE , TRUE  },
    { "--file"   , 2, CopyFile     , MAX_WAIT_TIME, TRUE , TRUE  },
    { "--mem"    , 1, SendBinary   , MIN_WAIT_TIME, FALSE, FALSE },
    { "--dir"    , 1, DrawDir      , MID_WAIT_TIME, TRUE,  TRUE  },
};
#define CHOICETSIZE (int)sizeof(choice_t)/(int)sizeof(choice_t[0])


/*
 * Excute l'opration
 */
static int ExecOperation(int choice)
{
    int drive ;

    if (choice_t[choice].mseconds == MID_WAIT_TIME)
        cc90.timeout = cc90.timeout_mid;
    else    
    if (choice_t[choice].mseconds == MAX_WAIT_TIME)
        cc90.timeout = cc90.timeout_max;
    else
        cc90.timeout = choice_t[choice].mseconds ;

#ifdef DEBUG_MODE
    const char typename[][8] = {"INVALID","TO","SAP","DCMOTO","FILE"} ;
    for(drive=0;drive<4;drive++)
        fprintf(fp_log,"%d. %8s DIR=%d drive=%d: diskname='%s' pattern='%s'\n",
                    drive,
                    typename[dk[drive].type],
                    dk[drive].dir,
                    dk[drive].drive,
                    dk[drive].dkname,
                    dk[drive].pattern
                    ) ;
    fprintf(fp_log,"choice = %d\n",choice) ;
#endif

    /* Filtrage des rsultats */
    for(drive=0;drive<4;drive+=2)
    {
        if (dk[drive].type != INVALID_TYPE)
        {
            if (dk[drive].pattern[0] == '\0')
                strcat(dk[drive].pattern,"*") ;

            /* Interdit les copies sur lui-mme */
            if ((dk[drive].type == dk[drive+1].type)
             && (dk[drive].drive == dk[drive+1].drive)
             && (strcmp(dk[drive].dkname,dk[drive+1].dkname) == 0)
             && (strcmp(dk[drive].pattern,dk[drive+1].pattern) == 0))
                return PError(dk[drive].arg,ERR_SAME_DISKS) ;
            /* Interdit les wildcards sur les chemins de fichier */
            if (strpbrk(dk[drive].dkname,"*?") != NULL)
                return PError(dk[drive].arg,ERR_ILLEGAL_WILDCARDS) ;
            if ((strpbrk(dk[drive+1].dkname,"*?") != NULL)
            /* Interdit les awildcards sur les destinations */
             || (strpbrk(dk[drive+1].pattern,"*?") != NULL))
                return PError(dk[drive+1].arg,ERR_ILLEGAL_WILDCARDS) ;
        }
    }
    
    /* Demande d'initialisation ct Thomson */
    cc90.diskop = choice_t[choice].initthomson ;

    /* Initialise les lecteurs si requis */
    if ((choice_t[choice].initdrives == TRUE)
     && (InitDrives() == FALSE))
        return FALSE ;

    /* Excution du programme */
    (*choice_t[choice].prog)() ;

    /* Fermeture ventuelle des dclarations */
    SerialClose() ;
    CloseFpFile() ;
    return TRUE ;
}


/*
 * Reprage des lments dans le chemin de fichier
 */
void /*@unused@*/
ParsePath(int drive, char *string) /*@globals errno@*/
{
    DIR *drbase = NULL ;
    char *found = NULL ;
    char *separ = NULL ;
    char *wildcard = NULL ;
    int index = 0 ;

    /* Sauve tout le chemin de fichier */
    (void)snprintf(dk[drive].arg,STRING_SIZE_MAX-1,"%s",string) ;

    /* Repre le numro de lecteur */
    dk[drive].external = FALSE ;
    dk[drive].drive = 0 ;
    if (isdigit((int)string[0]) != 0)
    {
        if ((strlen(string) == 1) || (string[1] == ':'))
        {
            dk[drive].drive = (UBYTE)(string[0] - '0') ;
            index = 1 + ((string[1] == ':') ? 1 : 0) ;
        }
        else
        if (((strlen(string) == 2) || (string[2] == ':'))
           && (toupper((int)string[1]) == (int)'B'))
        {
            dk[drive].drive = (UBYTE)(string[0] - '0') ;
            dk[drive].external = TRUE ;
            index = 2 + ((string[2] == ':') ? 1 : 0) ;
        }
    }

    /* Sauve le chemin de fichier sans le lecteur */
    (void)snprintf(dk[drive].dkname,STRING_SIZE_MAX-1,"%s",&string[index]) ;

    /* Ajoute un slash/backslash  la fin si rpertoire */
    if ((index == 0)
     && (dk[drive].dkname[0] != '\0')
     && ((drbase = opendir(dk[drive].dkname)) != NULL))
    {
        if ((dk[drive].dkname[strlen(dk[drive].dkname)-1] != '\\')
         && (dk[drive].dkname[strlen(dk[drive].dkname)-1] != '/'))
            strcat(dk[drive].dkname,SYSTEMSLASH) ;
        (void)closedir(drbase) ;
    }

    /* Dfinit le type de fichier */
    if ((found = strstr(dk[drive].dkname, ".sap")) != NULL)
        dk[drive].type = SAP_TYPE ;
    else
    if ((found = strstr(dk[drive].dkname, ".fd")) != NULL)
        dk[drive].type = DCMOTO_TYPE ;
    else
    if ((found = strstr(dk[drive].dkname, ".qd")) != NULL)
        dk[drive].type = DCMOTO_TYPE ;
    else
        dk[drive].type = (index != 0) ? TO_TYPE : FILE_TYPE ;

    /* Recherche si jokers
     * Si un joker se trouve avant un suffixe SAP, QD ou FD,
     * il s'agit d'un FILE_TYPE */
    if (found != NULL)
    {
        if (((wildcard = strpbrk(dk[drive].dkname,"*?")) != NULL)
         && (wildcard < found))
            dk[drive].type = FILE_TYPE ;
    }

    /* Recherche un '/' ou un '\' */
    if ((separ = strrchr(dk[drive].dkname,'/')) == NULL)
        separ = strrchr(dk[drive].dkname,'\\') ;

    /* Ajuste le chemin et le nom de fichier */
    if ((dk[drive].type == TO_TYPE)
     || ((separ == NULL) && (dk[drive].type == FILE_TYPE)))
    {
        (void)snprintf(dk[drive].pattern,STRING_SIZE_MAX-1,"%s",dk[drive].dkname) ;
        dk[drive].dkname[0] = '\0' ;
    }
    else
    {
        if ((separ != NULL) &&( found < separ))
        {
            (void)snprintf(dk[drive].pattern,STRING_SIZE_MAX-1,"%s",&separ[1]) ;
            separ[0] = '\0' ;
        }
    }

    /* Dtermine la nature de l'entre */
    dk[drive].dir = FALSE ;
    if ((dk[drive].pattern[0] == '\0')
     || (strpbrk(dk[drive].pattern,"*?")) != NULL)
        dk[drive].dir = TRUE ;

    /* Initialise le nom de disque si absent */
    if ((dk[drive].type != TO_TYPE)
     && (dk[drive].dkname[0] == '\0'))
        strcat(dk[drive].dkname,".") ;
}


/*
 * Rcupre le chemin des disquettes source et destination
 */
static int
TypeAnyDisk(int drive, char *message) /*@globals errno,stdout,stdin@*/
{
    char string[STRING_SIZE_MAX] ;
    (void)printf("   %s : ",message) ;
    (void)scanf(" %299s",string) ;
    ParsePath(drive,string) ;
    return (dk[drive].type == INVALID_TYPE) ? FALSE : TRUE ;
}


/*
 * Rcupre le chemin du lecteur SAP/FD/QD
 */
static int
TypeVirtualDisk(int drive) /*@globals errno,stdout,stdin@*/
{
    char string[STRING_SIZE_MAX] ;
    (void)printf("   Disquette SAP/FD/QD : ") ;
    (void)scanf(" %299s",string) ;
    ParsePath(drive,string) ;
    return ((dk[drive].type == SAP_TYPE)
         || (dk[drive].type == DCMOTO_TYPE)) ? TRUE : FALSE ;
}


/*
 * Rcupre l'entrelacement
 */
static int
TypeInterlace(void) /*@globals errno,stdout,stdin@*/
{
    (void)printf("   Entrelacement : ") ;
    (void)scanf(" %d",&opt.interlace) ;
    return ((opt.interlace >= 1) 
         && (opt.interlace <= 15)) ? TRUE : FALSE ;
}


/*
 * Rcupre l'entrelacement
 */
static void
TypeConvert(void) /*@globals errno,stdout,stdin@*/
{
    char string[3] ;
    do
    {
        (void)printf("   Conversion des fichiers (o/n)? ") ;
        (void)scanf(" %1s",string) ;
    } while ((toupper((int)string[0]) != (int)'O') && (toupper((int)string[0]) != (int)'N')) ;
    conv.convert = (toupper((int)string[0]) == (int)'O') ? TRUE : FALSE ;    
}


/*
 * Slecteur
 */
static void
Selector(void) /*@globals errno,stdout,stdin@*/
{
    int choice ;
    char string[3] ;
    int do_it ;
    int i ;
    int counter ;
    int validate_table[8] ;

    int choice_corresp[] = {
        CREATE_CHOICE,
        FORMAT_CHOICE,
        DELETE_CHOICE,
        DIR_CHOICE,
        FILE_CHOICE,
        DISK_CHOICE
    } ;

    PrintHelpHead() ;
    PrintAdviceHead() ;

    while(1 == 1)
    {
        (void)printf("--------------------------------------------\n") ;
        (void)printf("   1. Creer une disquette SAP/FD/QD\n") ;
        (void)printf("   2. Formater une disquette\n") ;
        (void)printf("   3. Effacer des fichiers Thomson\n") ;
        (void)printf("   4. Afficher un repertoire\n") ;
        (void)printf("   5. Copier des fichiers\n") ;
        (void)printf("   6. Copier une disquette\n") ;
        (void)printf("   Autre = SORTIE\n") ;
        (void)printf("Votre choix : ") ;
        (void)scanf(" %2s",string) ;
        /* Filtrage de la rponse */
        choice = (int)(string[0] - '0') ;
        if ((strlen(string) != 1)
         || (choice < 1)
         || (choice > ((int)sizeof(choice_corresp) / 4)))
            return ;

        InitStructures() ;
        do_it = FALSE ;

        switch(choice)
        {
             /* Cration de disquette */
             case 1 :
                 /* Rcupration du chemin de fichier */
                 if (TypeVirtualDisk(DRIVE0) == FALSE)
                 {
                     break ;
                 }

                 /* Affichage du menu selon le disque */
                 for(i=0;i<8;i++) validate_table[i] = FALSE ;
                 counter = 0 ;
                 for(i=1;i<8;i++)
                 {
                     if (((dk[DRIVE0].type == DCMOTO_TYPE)
                          && (strstr(dk[DRIVE0].dkname, ".qd") != NULL)
                          && (strcmp(cont[i].dcmotosuffix,"qd") == 0))
                      || ((dk[DRIVE0].type == DCMOTO_TYPE)
                          && (strstr(dk[DRIVE0].dkname, ".fd") != NULL)
                          && (strcmp(cont[i].dcmotosuffix,"fd") == 0))
                      || ((dk[DRIVE0].type == SAP_TYPE)
                          && (cont[i].sap != 0)))
                     {
                         (void)printf("   %d. Disque %d face", i, (int)cont[i].maxdrive) ;
                         if (cont[i].maxdrive > 1) (void)printf("s") ;
                         (void)printf(", %d pistes, ",(int)cont[i].nbtracks) ;
                         if (cont[i].secsize == 128)
                             (void)printf("simple") ;
                         else
                             (void)printf("double") ;
                         (void)printf(" densite\n") ;
                         validate_table[i] = TRUE ;
                         counter++ ;
                     }
                 }
                 /* Requte ventuelle avec filtrage du choix */
                 if (counter == 0) break ;
                 if (counter != 1)
                 {
                     (void)printf("   Votre choix : ") ;
                     (void)scanf(" %d",&i) ;
                 }
                 else i = 1 ;
                 if ((i > 7) || (validate_table[i] == FALSE))
                     break ;
                 /* Enregistrement du choix */
                 dk[DRIVE0].cont = i ;
                 do_it = TRUE ;
                 break ;

             /* Formatage de disquette */
             case 2 :
                 if (((do_it = TypeAnyDisk(DRIVE0,"Disquette a formater")) == TRUE)
                  && ((dk[DRIVE0].type == TO_TYPE) && (dk[DRIVE0].drive != 4)))
                     do_it = TypeInterlace() ;
                 break ;

             /* Effacement de fichiers */
             case 3 :
                 do_it = TypeAnyDisk(DRIVE0,"Fichier(s) a effacer") ;
                 break ;

             /* Affichage d'un rpertoire */
             case 4 :
                 do_it = TypeAnyDisk(DRIVE0,"Repertoire a afficher") ;
                 break ;

             /* Copie de fichiers */
             case 5 :
                 TypeConvert() ;
                 if ((do_it = TypeAnyDisk(DRIVE0,"Repertoire source")) == TRUE)
                     do_it = TypeAnyDisk(DRIVE1,"Repertoire destination") ;
                 break ;

             /* Copie de disquette */
             case 6 :
                 (void)printf("   Nombre de faces a copier (1-2) ? ") ;
                 (void)scanf(" %2s",string) ;
                 if ((strlen(string) != 1)
                  || (string[0] < '1')
                  || (string[0] > '2'))
                 {
                     do_it = FALSE ;
                     break ;
                 }
                 if ((do_it = TypeAnyDisk(DRIVE0,"Disquette source premiere face")) == TRUE)
                     do_it = TypeAnyDisk(DRIVE1,"Disquette destination premiere face") ;
                 if (string[0] == '2')
                     if ((do_it = TypeAnyDisk(DRIVE2,"Disquette source deuxieme face")) == TRUE)
                         do_it = TypeAnyDisk(DRIVE3,"Disquette destination deuxieme face") ;
                 break ;
         }
         /* Vide le tampon des scanf */
         do { i = getchar() ; } while ((i != (int)'\n') && (i != EOF)) ;
         /* Excute ventuellement le programme */
         if (do_it == TRUE)
             (void)ExecOperation(choice_corresp[choice-1]) ;
    }
}


/*
 * Vrifie si nombre dcimal
 */
int /*@unused@*/
CheckIfValidNumber(char *string)
{
    int i ;
    for(i=0;i<(int)strlen(string);i++)
        if (isdigit((int)string[i]) == 0)
            return FALSE ;
    return TRUE ;
}


/*
 * Reprage des formats de conversion
 */
int /*@unused@*/
InitConversionFormat(char *string) /*@globals errno@*/
{
    int i ;
    char *token ;
    const struct {
        int   /*@dependent@*/ *addr ;
        char  name[4] ;
        int   value ;
      } table[6] = {
        { &conv.sndtype, "SND", CONV_SND_SND },
        { &conv.sndtype, "SON", CONV_SND_SON },
        { &conv.txttype, "ASC", CONV_TXT_ASC },
        { &conv.txttype, "ASM", CONV_TXT_ASM },
        { &conv.maptype, "MAP", CONV_GFX_SNP },
        { &conv.maptype, "PPM", CONV_GFX_PPM }
    } ;
    int TSIZE = 6 ;

    token = strtok(string,"/") ;
    while(token)
    {
        i = 0 ;
        while((i < TSIZE) && (strcmp(token,table[i].name) != 0)) i++ ;
        if (i == TSIZE) return FALSE ;
        *table[i].addr = table[i].value ;
        token = strtok(NULL,"/") ;
    }
    return TRUE ;
}


/*
 * Analyse des arguments de la ligne de commande
 */
static void
ReadArgs(int argc, char *argv[]) /*@globals errno@*/
{
    int choice ;
    int argp ;
    int drive = 0 ;
    long long_number ;
    
    char string[STRING_SIZE_MAX] ;

    /* Initialise les structures du programme */
    InitStructures() ;

    /*
     * Lecture de la commande
     */
    choice = 0 ;
    while(strcmp(argv[1],choice_t[choice].command) != 0)
    {
        if (choice == CHOICETSIZE)
            PrintAdvice(argv[1], "Commande inconnue") ;
        choice++ ;
    }

    /*
     * Lecture des arguments
     */
    drive = 0 ;
    for(argp=2;argp<argc;argp++)
    {
        if (argv[argp][0] == '-')
        {
            switch ((int)strlen(argv[argp]))
            {
                case 2 :
                    switch(argv[argp][1])
                    {
                        case '0': conv.force40cols = TRUE ; break ;
                        case 'a': opt.autochg      = 0x80 ; break ;
                        case 'c': conv.convert     = TRUE ; break ;
                        case 'q': opt.quick        = TRUE ; break ;
                        case 'g': conv.gammac      = (conv.gammac == TRUE) ? FALSE : TRUE ; break ;
                        default:
                            PrintAdvice(argv[argp],"Option inconnue ou incomplete") ;
                            break ;
                    }
                    break ;
                default :
                    switch(argv[argp][1])
                    {
                        case 'b':
                            long_number = atoi(&argv[argp][2]) ;
                            if ((CheckIfValidNumber(&argv[argp][2]) == FALSE)
                             || (long_number < 0)
                             || (long_number > 30))
                                PrintAdvice(argv[argp],"Numero de banque de 0 a 30") ;
                            opt.bank = (UBYTE)long_number ;
                            break;

                        case 'c':
                            conv.convert = TRUE ;
                            strcpy(string,&argv[argp][2]) ;
                            if (InitConversionFormat(string) == FALSE)
                                PrintAdvice(string,"Valeur de conversion inconnue") ;                            
                            break;

                        case 'd':
                            dk[DRIVE0].cont = atoi(&argv[argp][2]) ;
                            if ((CheckIfValidNumber(&argv[argp][2]) == FALSE)
                             || (dk[DRIVE0].cont < 0)
                             || (dk[DRIVE0].cont > 7))
                                PrintAdvice(argv[argp],"Numero de controleur de 0 a 7") ;
                            break;

                        case 'e':
                            long_number = atoi(&argv[argp][2]) ;
                            if ((CheckIfValidNumber(&argv[argp][2]) == FALSE)
                             || (long_number < 1)
                             || (long_number > 15))
                                PrintAdvice(argv[argp],"Entrelacement de 1 a 15") ;
                            opt.interlace = (int)long_number ;
                            break;

                        case 's':
                            long_number = atoi(&argv[argp][2]) ;
                            if ((CheckIfValidNumber(&argv[argp][2]) == FALSE)
                             || (long_number <= 0))
                                PrintAdvice(argv[argp],"Erreur debit WAV") ;
                            conv.sample_rate = (ULONG)long_number ;
                            break;

                        default:
                            PrintAdvice(argv[argp],"Option inconnue ou incomplete") ;
                            break ;
                    }
            }
        }
        else
        {
            /* Limite le nombre d'arguments */
            if (drive == choice_t[choice].maxnbfiles)
                PrintAdvice(argv[argp],"Argument excedentaire") ;

            /* Repre les lments du chemin de fichier */
            ParsePath(drive,argv[argp]) ;

            /* Disquette suivante */
            drive++ ;
        }
    }
    (void)ExecOperation(choice) ;
}


/*
 * Programme principal
 */
int
main(int argc, char *argv[]) /*@globals errno,stdout,stdin@*/
{
    int argp ;

    /* Elague les arguments */
    for(argp=0;argp<argc;argp++)
        (void)Chop(argv[argp]) ;

#ifdef DEBUG_MODE
    fp_log = fopen("cc90.log","w") ;
    for(argp=0;argp<argc;argp++)
        fprintf(fp_log,"%s ",argv[argp]) ;
    fprintf(fp_log,"\n") ;
    fp_map = fopen("map.log","w") ;
#endif

    /* Initialise les buffers de pistes */
    (void)AllocTracks() ;
    
    /* Entre dans le programme */
    switch(argc)
    {
        case 1  : Selector() ; break ;
        default : ReadArgs(argc,argv) ; break ;
    }

    return CloseAll() ;
}
