/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Systme de fichiers Thomson
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <ctype.h>

#include "defs.h"

#define DIRSTRSIZE  80

extern struct CC90_T cc90 ;
extern struct DISK_T dk[4] ;
extern struct OPT_T opt ;

extern UBYTE *track_buffer ;
extern UBYTE *dos_buffer ;
extern UBYTE *file_buffer ;
extern ULONG file_size ;
extern void AllocFile(size_t) ;
extern void FreeFile(void) ;

extern int PError(char *, int) ;

/* Primitives disquette */
extern int InitDrive(int) ;
extern int ReadSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int SaveSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int ToGetSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int ToPutSectors(int, UBYTE, UBYTE, UBYTE) ;
extern int SendInfo(UBYTE, UBYTE, UBYTE, UBYTE, UBYTE, UBYTE) ;

/* Oprations PC */
static struct dirent /*@null@*/ *de = NULL ; /* Entre DIR */
extern FILE *fp_file ;
extern int OpenFpFile(int, char *, char *) ;
extern void CloseFpFile(void) ;
extern int LoadFpFile(char *) ;
extern char GetKeyboardHit(void) ;

extern const struct CONTROLLER cont[8] ;

static char pathname2[STRING_SIZE_MAX] ;  /* Seulement pour PathNameString() */


#ifdef DEBUG_MODE
    extern FILE *fp_log ;
#endif


/*
 * Lit la piste 20 d'une disquette
 *
 * Les secteurs de piste sont lus d'une traite par le Thomson
 * Ensuite :
 * Le secteur de nom (secteur 1) est transfr
 * Le secteur de FAT (secteur 2) est transfr
 * Les secteurs de catalogue sont transfrs un par
 * un jusqu' atteindre la fin du catalogue, soit
 * une entre commenant par 0xFF, soit le dernier
 * secteur (secteur 16)
 */
int ReadDirTrack(int drive)
{
    size_t i ;
    UBYTE sector ;
    size_t sector_index ;
    int reach_end ;

    if ((dos_buffer == NULL) || (track_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

    /* Efface le buffer de piste 20 */
    memset(&dos_buffer[drive * TRACKSIZE],0xFF,TRACKSIZE) ;

    dk[drive].fat = dos_buffer + (drive * TRACKSIZE) + ((FATSECTOR - 1) * dk[drive].secsize) ;
    switch (dk[drive].type)
    {
        case SAP_TYPE :
        case DCMOTO_TYPE :
            if (ReadSectors(drive, DIRTRACK, FIRSTB1, LASTB2) == FALSE)
                return FALSE ;
            memcpy(dos_buffer + (drive * TRACKSIZE), track_buffer, TRACKSIZE) ;
            if(*(dk[drive].fat++) != 0x00)
                return PError(dk[drive].arg,ERR_BAD_FAT) ;
            break ;

        case TO_TYPE :
            if (dk[drive].drive <= 3)
                if (SendInfo(CODE_LOADSECTORS, dk[drive].drive, DIRTRACK, FIRSTB1, LASTB2, 0) == FALSE)
                    return FALSE ;
            /* Charge les secteurs de nom et de FAT */
            if (ToGetSectors(drive, DIRTRACK, FIRSTB1, 3) == FALSE)
                return FALSE ; 
            memcpy(&dos_buffer[drive * TRACKSIZE], track_buffer, dk[drive].secsize * 3) ;
            if(*(dk[drive].fat++) != 0x00)
                return PError(dk[drive].arg,ERR_BAD_FAT) ;
            /* Charge les secteurs de catalogue */
            sector = 3 ;
            reach_end = FALSE ;
            while ((sector <= LASTB2) && (reach_end == FALSE))
            {
                /* Rcupre le secteur */
                if (sector != 3)
                    if (ToGetSectors(drive, DIRTRACK, sector, sector) == FALSE)
                        return FALSE ;
                sector_index = (sector - 1) * dk[drive].secsize ;
                memcpy(&dos_buffer[drive * TRACKSIZE] + sector_index,
                        (UBYTE/*@out@*/*)track_buffer + sector_index,
                       dk[drive].secsize) ;
                /* Vrifie si secteur de fin */
                for(i=0;i<dk[drive].secsize;i+=ENTRYSIZE)
                    if(*(&dos_buffer[drive * TRACKSIZE] + sector_index + i) == 0xFF)
                        reach_end = TRUE ;
                sector++ ;
            }
            break ;
    }
    return TRUE ;
}


/*
 * Sauve le secteur de FAT et de catalogue
 *
 * Les secteurs de FAT et de catalogue sont d'abord
 * transfrs vers le Thomson, puis sont sauvs
 * conscutivement
 * Ainsi, pour peu que les deux secteurs soient
 * suffisament loigns l'un de l'autre, ils 
 * seront sauvs en une seule passe
 */
static int SaveFatAndEntry(int drive)
{
    UBYTE entry_sector ;
    
    if ((dos_buffer == NULL) || (track_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

    entry_sector = (UBYTE)((dk[drive].entry - &dos_buffer[drive * TRACKSIZE]) / dk[drive].secsize) ;

    switch (dk[drive].type)
    {
        case SAP_TYPE :
        case DCMOTO_TYPE :
            memcpy(track_buffer,dos_buffer + (drive * TRACKSIZE),TRACKSIZE) ;
            return SaveSectors(drive, DIRTRACK, FIRSTB1, LASTB2);
        case TO_TYPE :
            memcpy(track_buffer + ((FATSECTOR - 1) * dk[drive].secsize),
                   dos_buffer + (drive * TRACKSIZE) + ((FATSECTOR - 1) * dk[drive].secsize),
                   dk[drive].secsize) ;
            if (ToPutSectors(drive, DIRTRACK, FATSECTOR, FATSECTOR) == FALSE)
                return FALSE ;
            memcpy(track_buffer + (entry_sector * dk[drive].secsize),
                   &dos_buffer[drive * TRACKSIZE] + (entry_sector * dk[drive].secsize),
                   dk[drive].secsize) ;
            if (ToPutSectors(drive, DIRTRACK, entry_sector + 1, entry_sector + 1) == FALSE)
                return FALSE ;
            if (dk[drive].drive <= 3)
            {
                if (SendInfo(CODE_SAVESECTORS,
                               dk[drive].drive,
                               DIRTRACK,
                               FATSECTOR,
                               FATSECTOR, 0) == FALSE)
                    return FALSE ;
                if (SendInfo(CODE_SAVESECTORS,
                               dk[drive].drive,
                               DIRTRACK,
                               entry_sector + 1,
                               entry_sector + 1, 0) == FALSE)
                    return FALSE ;
            }
            return TRUE ;
    }
    return PError(dk[drive].arg,ERR_BAD_FILE_FORMAT) ;
}


/*
 * Rcupration du secteur de FAT
 * Utilis exclusivement pour la copie de disquette
 */

int GetFat(int drive)
{
    size_t fatindex = (FATSECTOR - 1) * dk[drive].secsize ;

    if ((dos_buffer == NULL) || (track_buffer == NULL))
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

    dk[drive].fat = dos_buffer + (drive * TRACKSIZE) + fatindex ;

    /* Charge la FAT */
    if (ReadSectors(drive, DIRTRACK, FATSECTOR, FATSECTOR) == FALSE)
        return FALSE ;
    /* Copie la FAT dans le buffer */
    memcpy(dk[drive].fat, (UBYTE/*@out@*/*)track_buffer + fatindex, dk[drive].secsize);

    /* Vrifie la validit de la FAT */
    if(*(dk[drive].fat++) != 0x00)
        return PError(dk[drive].arg,ERR_BAD_FAT) ;

    return TRUE ;
}


/* ========================================================================== */
/*                      RECHERCHE DE L'ENTREE SUIVANTE                        */
/* ========================================================================== */


/*
 * Vrifie si le nom de fichier et le pattern correspondent
 */
static int ifmatch(char *pattern, char *filename)
{
    struct {
        char /*@temp@*/ *s1 ;
        char /*@temp@*/ *s2 ;
    }  lst[(strlen(pattern) > strlen(filename))
                ? strlen(pattern)+2
                : strlen(filename)+2] ;

    char /*@temp@*/ *s1 = pattern;
    char /*@temp@*/ *s2 = filename ;
    int i = 0, flg ;

    memset(&lst, 0, sizeof(lst)) ;

    if ((*s1 == '\0') || (*s2 == '\0')) return FALSE ;

    while(1 == 1)
    {
        flg = FALSE ;
        switch (*s1)
        {
            case '\0' : if (*s2 == '\0') return TRUE ;
                        if (s1 == pattern) return FALSE ;
                        if(*(s1-1) == '*') return TRUE ;
                        if(*(s1-1) != '?') return FALSE ;
                        flg = TRUE ;
                        break ;
            case '*' :  lst[i].s1 = s1++ ;
                        lst[i++].s2 = s2 ;
                        break ;
            case '?' :  s1++ ;
                        if(*(s2++) == '\0') return FALSE ;
                        break ;
            default :   if (*s2 == '\0') return FALSE ;
                        if (*(s1++) != *(s2++)) flg = TRUE ;
                        break ;
        }
        if (flg == TRUE)
        {
            do {
                if(i == 0) return FALSE ;
                s1 = lst[--i].s1 ;
                s2 = ++lst[i].s2 ;
            } while(*s2 == '\0') ;
        }
    }
    return FALSE ;
}


/*
 * Convertit un nom de fichier PC au format Thomson
 *
 * Le nom ne doit pas dpasser 8 caractres
 * Le suffixe ne doit pas dpasser 3 caractres
 */
int /*@unused@*/
CreateThomsonName(char *pcfilename, char *tofilename) /*@globals errno@*/
{
    char name0[15] = "" ;
    char name1[15] = "" ;
    char name2[15] = "" ;
    char *token ;

    strncpy(name0,pcfilename,14) ;
    strcpy(name1,(token = strtok(name0,".")) ? token : "\0") ;
    strcpy(name2,(token = strtok(NULL,".")) ? token : "\0") ;
    if ((strlen(name1) > 8) || (strlen(name2) > 3))
        return PError(pcfilename, ERR_BAD_FILE_NAME) ;
    (void)snprintf(tofilename,12,"%-8s%-3s",name1,name2) ;
    return TRUE ;
}


/*
 * Elimine les espaces de fin chane
 */
int /*@unused@*/
Chop(char *string)
{
    int i = (int)strlen(string) - 1 ;
    while ((i >= 0) && (string[i] <= '\x20')) i-- ;
    string[i+1] = '\0' ;
    return (int)strlen(string) ;
}


/*
 * Convertit un nom de fichier Thomson au format PC
 *
 * toentry  : pointeur sur l'entre catalogue du fichier Thomson
 * filename : pointeur sur le nom de fichier PC
 */
static int
CreatePcName(UBYTE *toentry, char *filename)
{
    char suffix[4] ;

    /* Ajoute le nom de fichier */
    if (toentry[0] == 0x00) return FALSE ;
    memcpy(filename,toentry,8) ;
    filename[8] = '\0' ;
    (void)Chop(filename) ;

    /* Ajoute ventuellement le suffixe */
    memcpy(suffix,&toentry[8],3) ;
    suffix[3] = '\0' ;
    if (Chop(suffix) != 0)
    {
        strcat(filename,".") ;
        strcat(filename,suffix) ;
    }

    /* Exclut les caractres interdits */
    if (strpbrk(filename," \"|/\\><*?")) return FALSE ;

    return TRUE ;
}


/*
 * Recherche fichier suivant pour Thomson
 *
 * TRUE               Entre valide
 * FALSE              Entre invalide
 * ERR_FILE_NOT_FOUND Fin de catalogue
 */
static int
ToFindNext(int drive)
{
    int match = FALSE ;

    if (dos_buffer == NULL)
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

    do
    {
        dk[drive].entry += ENTRYSIZE ;

        /* Sort si fin de catalogue physique */
        if (dk[drive].entry >= &dos_buffer[drive * TRACKSIZE] + (SECTORS_PER_TRACK * dk[drive].secsize))
            return ERR_FILE_NOT_FOUND ;
        
        /* Rcupre le nom de fichier */
        match = FALSE ;
        if (CreatePcName(dk[drive].entry,dk[drive].filename) == TRUE)
            match = ifmatch(dk[drive].pattern,dk[drive].filename) ;

        /* Repre la premire entre libre */
        if ((match == TRUE)
          || (((*dk[drive].entry == 0xFF)
           || (*dk[drive].entry == 0x00))
           && (dk[drive].free == NULL)))
            dk[drive].free = dk[drive].entry ;

        /* Sort si fin de catalogue logique */
        if (*dk[drive].entry == 0xFF)
            return ERR_FILE_NOT_FOUND ;

#ifdef DEBUG_MODE
        fprintf(fp_log,"ToFindNext %d ('%s') '%s'\n",
                           match,
                           dk[drive].pattern,
                           dk[drive].filename) ;
#endif
    } while(match == FALSE) ;

    return TRUE ;
}


/*
 * Recherche fichier suivant pour PC
 */
static int
PcFindNext(int drive) /*@globals errno@*/
{
    DIR *dirbase = NULL ;     /* Handler DIR pour reprage */
    char dirnametest[STRING_SIZE_MAX] ;
    dk[drive].filename[0] = '\0' ;

    if (dk[drive].drbase == NULL)
        return PError(dk[drive].dkname,ERR_IO_ERROR) ;

    while ((de = readdir(dk[drive].drbase)) != NULL)
    {
       if ((strcmp(de->d_name,".") != 0) && (strcmp(de->d_name,"..") != 0))
       {
            (void)snprintf(dk[drive].filename,STRING_SIZE_MAX-1,"%s",de->d_name) ;
            if (ifmatch(dk[drive].pattern,dk[drive].filename) == TRUE)
            {
                (void)snprintf(dirnametest,STRING_SIZE_MAX-1,"%s%s%s",dk[drive].dkname,SYSTEMSLASH,de->d_name) ;
                if ((dirbase = opendir(dirnametest)) == NULL)
                {
                    if (errno == ENOTDIR)
                        return TRUE ;
                }
                else
                {
                    (void)closedir(dirbase) ;
                    return ERR_FILE_IS_A_DIR ;
                }
            }
        }
    }
    return ERR_FILE_NOT_FOUND ;
}


/*
 * Recherche fichier suivant
 */
int
FindNext(int drive) /*@globals errno@*/
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return ToFindNext(drive) ;
        case FILE_TYPE :
            return PcFindNext(drive) ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== */
/*                     RECHERCHE DE LA PREMIERE ENTREE                        */
/* ========================================================================== */


/*
 * Recherche de la premire entre pour Thomson
 */
static int
ToFindFirst(int drive)
{
    if (dos_buffer == NULL)
        return PError("", ERR_NOT_ENOUGH_MEMORY) ;

#ifdef DEBUG_MODE
        fprintf(fp_log,"ToFindFirst\n") ;
#endif

    /* Initialise les repres */
    dk[drive].entry = dos_buffer + (drive * TRACKSIZE) + (2 * dk[drive].secsize) - ENTRYSIZE ;
    dk[drive].free = NULL ;

    return ToFindNext(drive) ;
}


/*
 * Recherche de la premire entre pour PC
 */
static
int PcFindFirst(int drive) /*@globals errno@*/
{
    if ((dk[drive].drbase = opendir(dk[drive].dkname)) == NULL)
    {
        switch(errno)
        {
            case ENOENT :
            case ENOTDIR :
                return PError(dk[drive].dkname,ERR_FILE_NOT_FOUND) ;
            case EACCES :
                return PError(dk[drive].dkname,ERR_PERMISSION_DENIED) ;
            default :
                return PError(dk[drive].dkname,ERR_CANT_OPEN_FILE) ;
        }
    }
    return PcFindNext(drive) ;
}


/*
 * Recherche premire entre
 */
int /*@unused@*/
FindFirst(int drive) /*@globals errno@*/
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return ToFindFirst(drive) ;
        case FILE_TYPE :
            return PcFindFirst(drive) ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== *
 *                     FERMETURE DE LECTURE DES ENTREES                       
 * ========================================================================== */


/*
 * Fermeture du rpertoire
 */
int
FindClose(int drive) /*@globals errno@*/
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return TRUE ;
        case FILE_TYPE :
            if (dk[drive].drbase)
            {
                (void)closedir(dk[drive].drbase) ;
                dk[drive].drbase = NULL ;
            }
            return TRUE ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== */
/*                   AFFICHAGE D'UNE LIGNE DE CATALOGUE                       */
/* ========================================================================== */


/*
 * Affiche une ligne de catalogue au format PC
 */
static int
PcDrawDirLine(int drive) /*@globals errno@,stdout*/
{
    long filesize ;
    char dirnametest[STRING_SIZE_MAX] ;

    if (de != NULL)
    {
        (void)snprintf(dirnametest,STRING_SIZE_MAX-1,"%s%s%s",dk[drive].dkname,SYSTEMSLASH,de->d_name) ;

        if (OpenFpFile(FALSE, dirnametest,"rb") == FALSE) return FALSE ;
        if (fseek(fp_file, 0, SEEK_END) != 0)
            return PError(dirnametest,ERR_IO_ERROR) ;
        if ((filesize = ftell(fp_file)) < 0)
            return PError(dirnametest,ERR_IO_ERROR) ;
        CloseFpFile() ;
        (void)printf("%-12s %ld\n",de->d_name,filesize) ;
    }

    return TRUE ;
}


/*
 * Rcupre la taille d'un fichier Thomson
 * Le programme en profitCreateThomsonName pour vrifier l'intgrit
 * du chanage des blocs
 */
static int
ThomsonFileSize(int drive, size_t *filesize)
{
    UBYTE block ;
    size_t blockcounter = 0 ;
    size_t lastsectorsize ;
    size_t sectorsize = (dk[drive].secsize == 256) ? (size_t)255 : dk[drive].secsize ;

    /* Rcupre le numro du premier bloc
     * Il ne doit pointer ni sur la piste de catalogue
     * ni  l'extrieur de la FAT */
    block = (UBYTE)dk[drive].entry[13] ;
    if (((block & 0xFE) == (DIRTRACK * 2))
     || (block >= (UBYTE)dk[drive].fatsize))
        return PError(dk[drive].filename,ERR_BAD_ENTRY) ;

    /* Calcule le nombre de blocs occups */
    while(block < 0xC0)
    {
        /* Rcupre le numro du bloc suivant
         * Il ne doit pas pointer sur la piste de catalogue,
         * et si il pointe  l'extrieur de la FAT, doit tre
         * compris entre 0xC1 et 0xC8
         * Le compteur de bloc ne doit pas dpasser la
         * capacit de la FAT */
        block = (UBYTE)dk[drive].fat[block] ;
        if (((block & 0xFE) == (DIRTRACK * 2))
         || ((block >= dk[drive].fatsize) && (block < 0xC1))
         || (block > 0xC8)
         || (blockcounter > (dk[drive].fatsize - 2)))
            return PError(dk[drive].arg,ERR_BAD_FAT) ;
        blockcounter++ ;
    }

    /* Rcupre la taille du dernier secteur
     * Elle ne doit pas tre suprieure  la taille d'un
     * secteur */
    lastsectorsize = (((UBYTE)dk[drive].entry[14] << 8) | (UBYTE)dk[drive].entry[15]) ;
    if (lastsectorsize > sectorsize)
        return PError(dk[drive].filename,ERR_BAD_ENTRY) ;

    /* Met  jour la taille de fichier */
    *filesize = ((((blockcounter - 1) << 3) + (block & 0x0F) - 1) * sectorsize) + lastsectorsize ;

    return TRUE ;
}


/*
 * Affiche une ligne de catalogue au format Thomson
 */
static int
ToDrawDirLine(int drive) /*@globals errno,stdout@*/
{
    int i ;
    int column ;
    size_t /*@unsignedintegraltype@*/ filesize = 0 ;
    char date[9] = "" ;
    char comment[9] = "" ;
    char typelist[] = "BDMA" ;
    char tofilename[13] = "" ;

    /* Rcupre le nom */
    if (CreateThomsonName(dk[drive].filename,tofilename) == FALSE)
        return FALSE ;
    for(i=0;i<11;i++)
        if (tofilename[i] < '\x20')
            tofilename[i] = '\x20' ;
    memmove(tofilename+9,tofilename+8,4) ;
    tofilename[8] = '\x20' ;

    /* Rcupre la taille */
    if (ThomsonFileSize(drive,&filesize) == FALSE)
        return FALSE ;

    /* Rcupre la date */
    if ((dk[drive].entry[24] >= 1)
     && (dk[drive].entry[24] <= 31)
     && (dk[drive].entry[25] >= 1)
     && (dk[drive].entry[25] <= 12))
        (void)snprintf(date,9,"%02d-%02d-%02d",
                     (int)dk[drive].entry[24],
                     (int)dk[drive].entry[25],
                     (int)dk[drive].entry[26]) ;

    /* Rcupre le commentaire */
    memcpy(comment,&dk[drive].entry[16],8) ;
    comment[8] = '\0' ;
    for(i=0;i<8;i++)
        if (comment[i] < '\x20')
            comment[i] = '\x20' ;

    /* Ecrit la ligne */
    (void)printf("  %-11s %c %c %6d %-8s %-8s [%02x]",
           tofilename,
           (dk[drive].entry[11] >= 4) ? '*' : typelist[(int)dk[drive].entry[11]],
           (dk[drive].entry[12] != 0) ? 'A' : 'B',
           (int)filesize,
           date,
           comment,
           (unsigned int)dk[drive].entry[13]) ;


    /* Ecrit les blocs occups */
    column = 48 ;
    i = (int)dk[drive].entry[13] ;
    while(i < 0xC0)
    {
        if (column > (DIRSTRSIZE - 5)) { (void)printf("\n%48s","") ; column = 48 ; } ;
        i = (int)dk[drive].fat[i] ;
        (void)printf(" %02x",(unsigned int)i) ;
        column += 4 ;
    }

    /* Ecrit la taille du dernier secteur */
    if (column > (DIRSTRSIZE - 10)) { (void)printf("\n%48s","") ; column = 48 ; } ;
    (void)printf(" (%02x%02x)\n",
           (unsigned int)dk[drive].entry[14],
           (unsigned int)dk[drive].entry[15]) ;

    return TRUE ;
}


/*
 * Affichage du rpertoire
 */
int
DrawDirLine(int drive) /*@globals errno,stdout@*/
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return ToDrawDirLine(drive) ;
        case FILE_TYPE :
            return PcDrawDirLine(drive) ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== */
/*                    AFFICHAGE DES INFOS DISQUETTE                           */
/* ========================================================================== */


/*
 * Affiche la ligne d'info d'une disquette Thomson
 */
static int
DrawToDiskLine(int drive)
{
    size_t i ;
    int diskfree = 0 ;
    char diskname[9] ;
    char densityname[][7] = { "Double", "Simple" } ;

    if (track_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;

    /* Rcupre le nom de disquette */
    memcpy(diskname,track_buffer,8) ;
    diskname[8] = '\0' ;
    for(i=0;i<8;i++)
        if (diskname[i] < '\x20')
            diskname[i] = '\x20' ;

    /* Rcupre le nombre de blocs libres sur le disque */
    for(i=0;i<dk[drive].fatsize;i++)
        if (dk[drive].fat[i] == 0xFF)
            diskfree++ ;

    /* Affiche la ligne */
    (void)printf("  %d:%-8s  %d occupes - %d libres - %s densite - Entrelacement %d\n",
             (int)dk[drive].drive,
             diskname,
             (int)(((dk[drive].nbtracks * 2) - diskfree) * dk[drive].secsize * (SECTORS_PER_TRACK / 2)),
             diskfree * (SECTORS_PER_TRACK / 2) * ((dk[drive].secsize == 256) ? 255 : dk[drive].secsize),
             densityname[(dk[drive].secsize == 256) ? 0 : 1],
             (int)(((cc90.lect & 0xF0) != 0) ? (cc90.lect & 0xF0) >> 4 : 1)) ;

    return TRUE ;
}


/*
 * Affichage des infos disquette
 */
int
DrawDiskLine(int drive)
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return DrawToDiskLine(drive) ;
        case FILE_TYPE :
            return TRUE ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== */
/*                           LECTURE DE FICHIER                               */
/* ========================================================================== */


/*
 * Charge un fichier au format Thomson
 */
static int
LoadThomsonFile(int drive)
{
    UBYTE block, nextblock ;
    ULONG size ;
    UBYTE sector ;
    ULONG datasize = (ULONG)((dk[drive].secsize == 256) ? (ULONG)255 : (ULONG)dk[drive].secsize) ;
    ULONG file_index = 0 ;
    UBYTE first = 0, last = 0 ;

    /* Alloue la mmoire pour le chargement du fichier
     * Au passage, l'intgrit du chanage des blocs est
     * vrifie */
    if (ThomsonFileSize(drive,&file_size) == FALSE)
        return FALSE ;
    AllocFile(file_size + MAXSECSIZE) ;
    if ((file_buffer == NULL) || (track_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;

    /* Charge le fichier */
    block = dk[drive].entry[13] ;

    while(block < 0xC0)
    {
        /* Dtermine le secteur de dpart et d'arrive
         * Le programme optimise le chargement en
         * slectionnant la piste entire quand 
         * la succession des blocs s'y prte */
        first = (((block & 1) != 0) ? FIRSTB2 : FIRSTB1) ;
        nextblock = dk[drive].fat[block] ;
        last = 0 ;
        if ((first == FIRSTB1) && (nextblock == block + 1))
        {
            nextblock = dk[drive].fat[nextblock] ;
            last = LASTB1 ;
        }
        last += first + ((nextblock < 0xC0) ? 7 : (nextblock & 0x0F) - 1) ;

        /* Charge les secteurs */
        if (ReadSectors(drive, block >> 1, first, last) == FALSE)
            return FALSE ;

        /* Ajuste taille et copie les donnes dans le buffer de fichier */
        for(sector=first;sector<=last;sector++)
        {
            memcpy(&file_buffer[file_index],
                   &track_buffer[dk[drive].secsize * (sector - 1)],
                   datasize);
            file_index += datasize ;
        }

        /* block suivant */
        block = nextblock ;
    }
    /* Traitement du dernier secteur */
    size = (dk[drive].entry[14] << 8) | dk[drive].entry[15] ;
    if (size > datasize)
        return PError(dk[drive].filename,ERR_BAD_ENTRY) ;
    if (size != 0)
        memcpy(&file_buffer[file_index], &track_buffer[dk[drive].secsize * last], size);

    return TRUE ;
}


/*
 * Charge un fichier au format PC
 */
static int
LoadPcFile(int drive)
{
    char pathname[STRING_SIZE_MAX] ;
    (void)snprintf(pathname,STRING_SIZE_MAX-1,"%s%s%s",dk[drive].dkname,SYSTEMSLASH,dk[drive].filename) ;
    return LoadFpFile(pathname) ;
}


/*
 * Chargement d'un fichier
 */
int
LoadFile(int drive)
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return LoadThomsonFile(drive) ;
        case FILE_TYPE :
            return LoadPcFile(drive) ;
        default :
            return PError(dk[drive].filename,ERR_CANT_OPEN_FILE) ;
    }
}


/* ========================================================================== */
/*                        SAUVEGARDE DE FICHIER                               */
/* ========================================================================== */

/*
 * Recherche un bloc libre dans la FAT
 *
 * le programme recherche le bloc libre le plus
 * proche du bloc de rfrence 'startblock',
 * prfrentiellement un bloc pair - pour occuper
 * une piste entire au final - et vers la
 * piste 0 - puisque c'est vers la piste 0 que se
 * trouve le plus petit nombre de blocs dans
 * la pluspart des cas -, ce qui vite  la tte
 * du lecteur de trop se dplacer
 * Pour le premier bloc d'un fichier, 'startblock'
 * doit tre le numro de bloc de la piste de
 * catalogue (DIRTRACK * 2).
 * Pour les blocs suivants, 'startblock' est le
 * numro du bloc courant
 */
static UBYTE
FindANewBlock(int drive, UBYTE startblock)
{
    int i ;
    int index = 0 ;
    int block ;

    while(index < (int)dk[drive].fatsize)
    {
        for(i=0;i<4;i++)
        {
            index = -index ;
            block = (int)((startblock & 0xFE) + index + ((UBYTE)i >> 1)) ;
            /* Ne renvoie le block que s'il se trouve dans
             * l'espace de la FAT et que la place est libre (=0xFF) */
            if ((block >= 0)
             && (block < (int)dk[drive].fatsize)
             && (dk[drive].fat[block] == 0xFF))
            {
                /* Met la FAT  jour si pas premier bloc */
                if ((startblock & 0xFE) != (DIRTRACK * 2))
                    dk[drive].fat[startblock] = (UBYTE)block ;
                /* Rserve la place trouve */
                dk[drive].fat[block] = 0xC0 ;
                return (UBYTE)block ;
            }
        }
       index += 2 ;
    }
    /* FAT pleine = erreur */
    return 0xFF ;
}


/*
 * Cre la chane de caractre du chemin
 */
char /*@unused@*//*@temp@*/
*PathNameString(int space_flag, int drive)
{
   char drivestring[3] = "" ;
   char rootstring[STRING_SIZE_MAX] = "" ;

   /* Numro de lecteur */
   if ((dk[drive].type == DCMOTO_TYPE)
    || (dk[drive].type == TO_TYPE))
      (void)snprintf(drivestring,3,"%d:",(int)dk[drive].drive) ;
   /* Nom de disquette */
   if ((dk[drive].type == DCMOTO_TYPE)
    || (dk[drive].type == SAP_TYPE)
    || (dk[drive].type == FILE_TYPE))
      (void)snprintf(rootstring,STRING_SIZE_MAX-1,"%s%s",dk[drive].dkname,SYSTEMSLASH) ;

   if (space_flag != 0)
       (void)snprintf(pathname2,STRING_SIZE_MAX-1,"%s%s%-12s",
                                              drivestring,
                                              rootstring,
                                              dk[drive].filename) ;
   else
       (void)snprintf(pathname2,STRING_SIZE_MAX-1,"%s%s%s",
                                              drivestring,
                                              rootstring,
                                              dk[drive].filename) ;
   return pathname2 ;
}


/*
 * Requte d'crasement/effacement de fichier
 */
int
AskRequest(int drive, int flag)
{
    char c ;

    while(cc90.yes == FALSE)
    {
        switch(flag)
        {
            case OVERWRITE_FLAG : (void)printf("   Ecraser") ; break ;
            case DELETE_FLAG    : (void)printf("   Effacer") ; break ;
            default : cc90.exit = TRUE ; return FALSE ;
        }
        (void)printf(" '%s' (o=oui/n=non/t=tous/s=sortie) ? ",PathNameString(FALSE, drive)) ;
        c = GetKeyboardHit() ;
        (void)printf("%c\n",c) ;

        switch((char)toupper((int)c))
        {
            case 'O' : return TRUE ;
            case 'T' : cc90.yes = TRUE ; return TRUE ;
            case 'N' : return FALSE ;
            case 'S' : cc90.exit = TRUE ; return FALSE ;
        }
    }
    return TRUE ;
}


/*
 * Effacement de l'entre Thomson et du chanage de blocs dans la FAT
 */
int /*@unused@*/
DeleteThomsonEntry(int drive)
{
    UBYTE block, nextblock ;

    /* Efface l'entre dans le catalogue */
    dk[drive].entry[0] = 0x00 ;

    /* Efface les blocs dans la FAT */
    block = dk[drive].entry[13] ;
    while(block <= 0xC0)
    {
        nextblock = dk[drive].fat[block] ;
        dk[drive].fat[block] = 0xFF ;
        block = nextblock ;
    }
    /* Sauvegarde de la FAT et du catalogue */
    return SaveFatAndEntry(drive) ;
}


/*
 * Sauve un fichier au format Thomson
 */
static int
SaveThomsonFile(int drive) /*@globals errno@*/
{
    const struct {
        char  suf[4] ;
        UBYTE type ;
        UBYTE flag ;
    } sufflist[] = {
        /* Basic Binaire */
        { "BAS", 0x00, 0x00 },
        { "BAT", 0x00, 0x00 },
        { "   ", 0x00, 0x00 },
        /* Basic Ascii */
        { "ABC", 0x00, 0xFF },
        { "CMP", 0x00, 0xFF },
        { "cmp", 0x00, 0xFF },
        { "pAM", 0x00, 0xFF },
        /* Data Ascii */
        { "DAT", 0x01, 0xFF },
        { "ASC", 0x01, 0xFF },
        { "CHG", 0x01, 0xFF },
        { "CAR", 0x01, 0xFF },
        /* Machine Binaire */
        { "BIN", 0x02, 0x00 },
        { "MAP", 0x02, 0x00 },
        { "PPM", 0x02, 0x00 },
        { "IMP", 0x02, 0x00 },
        { "SON", 0x02, 0x00 },
        { "SND", 0x02, 0x00 },
        { "PUK", 0x02, 0x00 },
        { "PAN", 0x02, 0x00 },
        { "POL", 0x02, 0x00 },
        { "QLP", 0x02, 0x00 },
        { "MO6", 0x02, 0x00 },
        /* Assembleur Ascii */
        { "ASM", 0x03, 0xFF },
        /* Exotique */
        { "PAR", 0x0A, 0x00 },
        { "DOC", 0x0A, 0x00 }
    };
#   define SUFFSIZE (int)sizeof(sufflist)/(int)sizeof(sufflist[0])

    time_t x;
    struct tm *t;

    ULONG i ;
    UBYTE sector ;
    UBYTE block, nextblock ;
    ULONG size ;
    ULONG datasize = (ULONG)((dk[drive].secsize == 256) ? 255 : dk[drive].secsize) ;
    ULONG file_index = 0L ;
    UBYTE firstsector = 0L, lastsector = 0L ;
    ULONG /*@unsignedintegraltype@*/ filesize = 0 ;
    char tofilename[12] = "" ;

    if ((file_buffer == NULL) || (track_buffer == NULL))
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;

    /* Reprage d'une entre Thomson valide */
    switch(FindFirst(drive))
    {
        case ERR_FILE_NOT_FOUND :
            /* Vrifie si entre libre dans le catalogue */
            if ((dk[drive].entry = dk[drive].free) == NULL)
                return PError(dk[drive].arg,ERR_DIRECTORY_FULL) ;
            /* Ecrit le nom de fichier */
            strcpy(dk[drive].filename,dk[drive].pattern) ;
            break ;

        case TRUE :
            /* Vrification de l'intgrit du fichier */
            if (ThomsonFileSize(drive,&filesize) == FALSE)
                return FALSE ;
            /* Effacement de l'entre et du chanage de blocs dans la FAT */
            if (DeleteThomsonEntry(drive) == FALSE)
                return FALSE ;
            break ;

        case FALSE :
            return FALSE ;
    }

    /* Initialise l'entre */
    if (dk[drive-1].type == FILE_TYPE)
        memset(dk[drive].entry,0x00,ENTRYSIZE) ;
    else
        memcpy(dk[drive].entry,dk[drive-1].entry, ENTRYSIZE) ;

    /* Cre le nom au fomat Thomson */
    if (CreateThomsonName(dk[drive].filename,(char/*@out@*/*)tofilename) == FALSE)
        return FALSE ;

#ifdef DEBUG_MODE
    fprintf(fp_log,"ToSavefile '%s' '%s' '%s'\n",dk[drive].pattern,dk[drive].filename,tofilename) ;
#endif
    /* Actualisation conditionnelle de l'entre */
    if (dk[drive].entry[0] == 0x00)
    {
        /* Actualisation du type et du flag de fichier */
        dk[drive].entry[11] = 0x00 ;
        dk[drive].entry[12] = 0x00 ;
        for(i=0;i<(ULONG)SUFFSIZE;i++) {
            if (strncmp(&tofilename[8],sufflist[i].suf,3) == 0) {
                dk[drive].entry[11] = (UBYTE)sufflist[i].type ;
                dk[drive].entry[12] = sufflist[i].flag ;
            }
        }

        /* Ajuste flags de fichier si Basic Ascii */
        if ((strncmp(&tofilename[8],"BAS",3) == 0)
         && (file_size > 1)
         && (file_buffer[0] == 0x0D)
         && (file_buffer[1] == 0x0A)) {
            dk[drive].entry[11] = 0x00 ;
            dk[drive].entry[12] = 0xFF ;
        }

        /* Actualisation du flag de lancement et espace mmoire si CHG */
        if (strncmp(&tofilename[8],"CHG",3) == 0)
        {
            printf("%d\n",file_size);
            dk[drive].entry[30] = opt.autochg | (((file_size+16384-1)/16384) & 0x3F) ;
        }

        /* Actualisation du checksum de nom si CHG */
        if(strncmp(&tofilename[8],"CHG",3) == 0)
        {
            dk[drive].entry[31] = 0x00 ;
            for(i=0;i<8;i++)
                dk[drive].entry[31] += (UBYTE)tofilename[i] ;
        }

    }

    /* Inscrit le nom de fichier */
    memcpy(dk[drive].entry,tofilename,11) ;

    /* Inscrit ventuellement la date */
    if ((dk[drive].entry[24] < 1) || (dk[drive].entry[24] > 31)) {
        (void)time(&x);
        t = gmtime(&x);
        if (t != NULL) {
            dk[drive].entry[24] = (UBYTE)t->tm_mday ;
            dk[drive].entry[25] = (UBYTE)t->tm_mon + 1 ;
            dk[drive].entry[26] = (UBYTE)(t->tm_year % 100) ;
        }
    }

    /* Inscrit les blocs de fichiers
     *
     * L'allocation des blocs de fichiers ncessaires
     *  la sauvegarde se fait en une seule fois pour
     * tre prvenu d'un dpassement de la capacit
     * de la disquette avant que quoique ce soit y
     * soit crit */

    /* Inscrit le premier bloc de fichier */
    if ((block = FindANewBlock(drive,DIRTRACK * 2)) == 0xFF)
        return PError(dk[drive].arg,ERR_FATFULL) ;
    dk[drive].entry[13] = block ;

    /* Inscrit les blocs occups dans la FAT */
    for(i=file_size; i>(datasize * 8); i-=(datasize * 8))
        if ((block = FindANewBlock(drive,block)) == 0xFF)
            return PError(dk[drive].arg,ERR_FATFULL) ;

    /* Au moins un secteur doit tre occup
       mme si la taille du fichier est nulle */
    dk[drive].fat[block] = 0xC0 + ((file_size != 0) ? 0 : 1) + ((i + datasize) / datasize);
    dk[drive].entry[14] = (i % datasize) >> 8 ;
    dk[drive].entry[15] = i % datasize ;

    /* Sauve le fichier */
    block = dk[drive].entry[13] ;
    while (block < 0xC0)
    {
        firstsector = ((block & 1) != 0) ? FIRSTB2 : FIRSTB1 ;
        lastsector  = firstsector + 7 ;
        nextblock = dk[drive].fat[block] ;

        /* Si fin de fichier */
        if (nextblock > 0xC0)
            lastsector = firstsector + (nextblock & 0x0F) - 1 ;
        else
        if (((block & 1) == 0) && (nextblock == (block + 1)))
        {
            block = nextblock ;
            nextblock = dk[drive].fat[block] ;
            lastsector = (nextblock > 0xC0)
                         ? firstsector + (nextblock & 0x0F) - 1 + 8
                         : LASTB2 ;
        }

        /* Ecrit le cluster dans la piste */
        for(sector=firstsector; sector<=lastsector; sector++)
        {
            size = ((nextblock > 0xC0) && (sector == lastsector))
                    ? (ULONG)((dk[drive].entry[14] << 8) | dk[drive].entry[15])
                    : datasize ;

            if (size != 0)
                memcpy(&track_buffer[(sector - 1) * dk[drive].secsize],
                       &file_buffer[file_index],
                       size );

            file_index += size ;
        }

        /* Sauve le cluster et cluster suivant */
        if (SaveSectors(1, block >> 1, firstsector, lastsector) == FALSE)
            return FALSE ;
        block = nextblock ;
    }

    /* Sauvegarde de la FAT et du catalogue */
    return SaveFatAndEntry(drive) ;
}


/*
 * Sauve un fichier au format PC
 */
static int
SavePcFile(int drive) /*@globals errno@*/
{
    char pathname[STRING_SIZE_MAX] ;

    if (file_buffer == NULL)
        return PError("",ERR_NOT_ENOUGH_MEMORY) ;

    strcpy(dk[drive].filename,dk[drive].pattern) ;
    (void)snprintf(pathname,STRING_SIZE_MAX-1,"%s%s%s",dk[drive].dkname,SYSTEMSLASH,dk[drive].filename) ;

    switch(OpenFpFile(TRUE, pathname,"rb"))
    {
        case FALSE :
            return FALSE ;
        case ERR_FILE_NOT_FOUND :
            break ;
        case TRUE :
            CloseFpFile() ;
            break ;
    }

    if (OpenFpFile(TRUE,pathname,"wb") == FALSE) return FALSE ;
    (void)fwrite(file_buffer, sizeof(char), file_size, fp_file) ;
    CloseFpFile() ;
    FreeFile() ;
    return TRUE ;
}


/*
 * Sauvegarde d'un fichier
 */
int
SaveFile(int drive) /*@globals errno@*/
{
    switch(dk[drive].type)
    {
        case TO_TYPE :
        case SAP_TYPE :
        case DCMOTO_TYPE :
            return SaveThomsonFile(drive) ;
        case FILE_TYPE :
            return SavePcFile(drive) ;
    }
    return PError(dk[drive].arg,ERR_CANT_OPEN_FILE) ;
}


/*
 * Initialise la piste 20
 */
int InitTrack20(int drive, int controller)
{
    size_t pos = cont[controller].secsize ;

    /* Remplissage de la piste */
    memset(track_buffer, 0xFF, TRACKSIZE) ;
    memset(&track_buffer[pos], 0xFE, cont[controller].secsize) ;
    memset(&track_buffer[pos], 0xFF, cont[controller].fatsize + 1) ;
    memset(&track_buffer[pos+41], 0xFE, 2) ;
    track_buffer[pos] = 0x00 ;

    /* Met  jour les registres contrleur */
    dk[drive].cont     = controller ;
    dk[drive].nbtracks = cont[controller].nbtracks ;
    dk[drive].fatsize  = cont[controller].fatsize  ;
    dk[drive].secsize  = cont[controller].secsize  ;

    /* Ecrit les secteurs de piste */
    return SaveSectors(drive, DIRTRACK, FIRSTB1, LASTB2) ;
}

