/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Primitives srie pour systmes Unix
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sysexits.h>
#include <sys/time.h>

#ifndef TRUE
#   define TRUE 1
#endif

#ifndef FALSE
#   define FALSE 0
#endif

static struct termios gOriginalTTYAttrs;
static int fileDescriptor = -1 ;
static int baudrate = 5 ;
static struct {
    long code ;
    long cycles ;
} baudtable[8] = {
     {   B110, 9077 },
     {   B300, 3328 },
     {   B600, 1664 },
     {  B1200,  832 },
     {  B2400,  416 },
     {  B4800,  208 },
     {  B9600,  104 },
     { B19200,   52 }
} ;


/*
 * Ferme le port srie
 */
void
SerialClose(void) /*@globals errno@*/
{
    if (fileDescriptor != -1)
    {
        /* Flushe le buffer */
        (void)tcdrain(fileDescriptor) ;

        /* Rtablit l'ancien port */
        (void)tcsetattr(fileDescriptor, TCSANOW, &gOriginalTTYAttrs) ;

        /* Ferme le port */
        (void)close(fileDescriptor);
    }
    fileDescriptor = -1 ;
}


/*
 * Ouvre le port srie
 *
 * portname = String du nom du port srie  ouvrir
 * bauds    = 1  110 bauds
 *            2  300 bauds
 *            3  600 bauds
 *            4  1200 bauds
 *            5  2400 bauds
 *            6  4800 bauds
 *            7  9600 bauds
 *            8  19200 bauds
 * datasize = 7  7 bits par donne
 *            8  8 bits par donnes
 */
int
SerialOpen(char *portname, int bauds, int datasize) /*@globals errno@*/
{
    struct termios  options;

    /* Sortie si dj ouvert */
    if (fileDescriptor != -1)
        return TRUE;

    /* Limitation du nombre de bauds */
    if ((bauds < 1) || (bauds > 8))
        return FALSE ;
    baudrate = bauds-1 ;
    
    /* Limitation de la taille d'une donne */
    if ((datasize < 7) || (datasize > 8))
        return FALSE ;

    /* Ouverture du port srie */
    if ((fileDescriptor = open(portname, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0)
        return FALSE;

    /* Rservation du port srie pour un usage exclusif */
    if (ioctl(fileDescriptor, TIOCEXCL) < 0)
        return FALSE;

    /* Demande le bloquage lors de l'usage */
    if (fcntl(fileDescriptor, F_SETFL, NULL) < 0)
        return FALSE;

    /* Sauvegarde options courantes pour restitution */
    if (tcgetattr(fileDescriptor, &gOriginalTTYAttrs) < 0)
        return FALSE;

    /* Programmation du port srie */
    options = gOriginalTTYAttrs;
    cfmakeraw(&options);
    options.c_cc[VTIME] = 0 ;  /* pas de timeout */
    options.c_cc[VMIN] = 0 ;    /* pas de timeout */
    (void)cfsetispeed(&options,baudtable[baudrate].code) ;
    (void)cfsetospeed(&options,baudtable[baudrate].code) ;
    options.c_cflag &= ~(PARENB | CSTOPB) ;
    options.c_cflag |= (CSIZE | ((datasize == 7) ? CS7 : CS8) | CRTSCTS | CREAD | CLOCAL) ;
    options.c_iflag &= ~(IXON | IXOFF | IXANY) ;
    options.c_iflag |= (IGNBRK | IGNPAR) ;
    options.c_oflag &= ~(OPOST) ;
    options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG) ;

    /* Activation du port srie */
    if (tcsetattr(fileDescriptor, TCSANOW, &options) < 0)
        return FALSE;

    /* Attente de suret */
    usleep(600) ;

    return TRUE ;
}


/*
 * Attend le nombre de cycles ncessaires  l'absorbtion de la donne
 */
static void
WaitCycles(void)
{
    long tl_sec, tl_usec ;
    static struct timeval tv;
    static struct timezone tz;

    gettimeofday(&tv, &tz);
    tl_sec = tv.tv_sec ;
    tl_usec = tv.tv_usec + (baudtable[baudrate].cycles * 10) ;
    while ((((tv.tv_sec-tl_sec)*1000000) + tv.tv_usec) <= tl_usec)
        gettimeofday(&tv, &tz);
}


/*
 * Envoie le bloc de donnes asynchrones sur le port srie
 */
int
SerialWrite(char c) /*@globals errno@*/
{
    /* Envoie la donne au Thomson */
    switch (write(fileDescriptor, &c, 1))
    {
        case 1 : WaitCycles() ; return 1 ;
        case 0 : return 0 ;
    }
    return -1 ;
}


/*
 * Reoit un bloc de donnes asynchrones du port srie
 */
char
SerialRead(void) /*@globals errno@*/
{
    char c ;
    /* Le retour de 'read' ne doit pas tre :
     * <0 : une erreur de lecture est survenue
     * =0 : le timeout est dpass */

    /* Envoie la donne au Thomson */
    switch (read(fileDescriptor, &c, 1))
    {
        case 1 : return c ;
        case 0 : return '\0' ;
    }
    return (char)-1 ;
}
