/*       ___   ___   ___   ___
 *      /_  \ /   \ /  _\ /  _\
 *     // \_/// \_/_\ / \\  / \\
 *    //  _ //  _ / \\\_// /  //
 *    \\_/ \\\_/ \\ //  /  \_//
 *     \___/ \___/ \\__/ \___/
 *     Transferts srie pour Thomson
 *
 *  Primitives srie pour Windows
 *
 *  Copyright (C) 2015 Puls
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <windows.h>
#include <assert.h>
#include <winioctl.h>

#ifndef TRUE
#   define TRUE 1
#endif

#ifndef FALSE
#   define FALSE 0
#endif

/* La taille des tampons de rception et d'mission
 * doivent tre pairs pour la compatibilit avec
 * Windows NT, 2000, XP et Vista
 */
#define rdSize  1200    /* Taille du tampon de rception */
#define txSize  1200    /* Taille du tampon d'mission */

static HANDLE hCom = NULL ;
static int baudrate = 5 ;


/*
 * Ferme le port srie
 */
void SerialClose(void)
{
    if (hCom)
    {
        CloseHandle(hCom) ;
        hCom = NULL ;
    }
}


/*
 * Ouvre le port srie
 *
 * portname = String du nom du port srie  ouvrir
 * bauds    = 1  110 bauds
 *            2  300 bauds
 *            3  600 bauds
 *            4  1200 bauds
 *            5  2400 bauds
 *            6  4800 bauds
 *            7  9600 bauds
 *            8  19200 bauds
 * datasize = 7  7 bits par donne
 *            8  8 bits par donnes
 */
int SerialOpen(char *portname, int bauds, int datasize)
{
    int baudtable[8] = {
        CBR_110,
        CBR_300,
        CBR_600,
        CBR_1200,
        CBR_2400,
        CBR_4800,
        CBR_9600,
        CBR_19200
    } ;
    DCB dcb ;
    COMMTIMEOUTS cto ;

    /* Sortie si dj ouvert */
    if(hCom) return TRUE ;

    /* Cre le port */
    hCom = CreateFile(portname,
                    GENERIC_READ | GENERIC_WRITE,
                    0,
                    0,
                    OPEN_EXISTING,
                    FILE_FLAG_OVERLAPPED,
                    0) ;

    if(!hCom) return FALSE ;

    /* Limitation du nombre de bauds */
    if ((bauds < 1) || (bauds > 8))
        return FALSE ;
    baudrate = bauds ;
    
    /* Limitation de la taille d'une donne */
    if ((datasize < 7) || (datasize > 8))
        return FALSE ;

    /* Dfinit les paramtres du port
     * Le port est paramtr et on attend un peu
     *   que la fonction ait fini son travail
     */
    memset(&dcb, 0, sizeof(DCB));
    dcb.DCBlength = sizeof(DCB) ;
    if ((!GetCommState(hCom, &dcb)))
        return FALSE ;
    dcb.BaudRate          = baudtable[baudrate-1] ;
    dcb.ByteSize          = datasize ;
    dcb.StopBits          = ONESTOPBIT ;
    dcb.fParity           = NOPARITY ;
    dcb.fBinary           = TRUE ;
    dcb.fDtrControl       = DTR_CONTROL_ENABLE;
    dcb.fOutxDsrFlow      = FALSE;
    dcb.fDsrSensitivity   = FALSE;
    dcb.fTXContinueOnXoff = FALSE;
    dcb.fOutX             = FALSE;
    dcb.fInX              = FALSE;
    dcb.fNull             = FALSE;
    dcb.fAbortOnError     = FALSE;
    dcb.fOutxCtsFlow      = TRUE;
    dcb.fRtsControl       = RTS_CONTROL_HANDSHAKE;

    if ((!SetCommState(hCom, &dcb)))
        return FALSE ;
    Sleep(60) ;

    /* Dfinit les timeouts
     *   Les timeouts sont paramtrs et on attend un peu
     *   que la fonction ait fini son travail
     */
    cto.ReadIntervalTimeout         = MAXDWORD ;
    cto.ReadTotalTimeoutMultiplier  = 0 ;
    cto.ReadTotalTimeoutConstant    = 0 ;
    cto.WriteTotalTimeoutMultiplier = 0 ;
    cto.WriteTotalTimeoutConstant   = 0 ;
    if ((!SetCommTimeouts(hCom, &cto)))
        return FALSE ;
    Sleep(60) ;

    /* Dfinit les tailles des buffers tampon */
    if (!SetupComm(hCom, rdSize, txSize))
        return FALSE ;
    Sleep(60) ;

    /* Vide les buffers tampon */
    PurgeComm(hCom,
                PURGE_TXABORT |
                PURGE_RXABORT |
                PURGE_TXCLEAR |
                PURGE_RXCLEAR
                ) ;
    Sleep(60) ;
    return TRUE ;
}


/*
 * Envoie le bloc de donnes asynchrones sur le port srie
 */
int SerialWrite(char c)
{
    DWORD nbytes = 0 ;
    OVERLAPPED o ;

    /* Cre l'event d'mission pour le mode asynchrone */
    memset(&o, 0, sizeof(OVERLAPPED)) ;
    if ((o.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL)
        return FALSE ;

    /* Envoie le bloc de donnes */
    if ((WriteFile(hCom, &c, 1, &nbytes, &o) == FALSE)
     && (GetLastError() == ERROR_IO_PENDING)
     && (WaitForSingleObject(o.hEvent, 0) == WAIT_OBJECT_0))
        GetOverlappedResult(hCom, &o, &nbytes, FALSE) ;

    /* Referme l'event d'mission */
    CloseHandle(o.hEvent) ;

    return (nbytes == 1) ? 1 : 0 ;
}


/*
 * Reoit un bloc de donnes asynchrones du port srie
 */
char SerialRead(void)
{
    DWORD nbytes = 0 ;
    OVERLAPPED o ;
    char c ;

    /* Cre l'event de rception pour le mode asynchrone */
    memset(&o, 0, sizeof(OVERLAPPED)) ;
    if ((o.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL)) == NULL)
        return FALSE ;

    /* Reoie le bloc de donnes */
    if ((ReadFile(hCom, &c, 1, &nbytes, &o) == FALSE)
     && (GetLastError() == ERROR_IO_PENDING)
     && (WaitForSingleObject(o.hEvent, 0) == WAIT_OBJECT_0))
        GetOverlappedResult(hCom, &o, &nbytes, FALSE) ;

    /* Referme l'event de rception */
    CloseHandle(o.hEvent) ;

    return (nbytes == 1) ? c : 0 ;
}
